"""Show materialized configuration of tox environments."""

from __future__ import annotations

import os
from textwrap import indent
from typing import TYPE_CHECKING

from colorama import Fore

from tox.config.loader.stringify import stringify
from tox.plugin import impl
from tox.session.cmd.run.common import env_run_create_flags
from tox.session.env_select import CliEnv, register_env_select_flags

if TYPE_CHECKING:
    from collections.abc import Iterable

    from tox.config.cli.parser import ToxParser
    from tox.config.sets import ConfigSet
    from tox.session.state import State
    from tox.tox_env.api import ToxEnv


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("config", ["c"], "show tox configuration", show_config)
    our.add_argument(
        "-k",
        nargs="+",
        help="list just configuration keys specified",
        dest="list_keys_only",
        default=[],
        metavar="key",
    )
    our.add_argument(
        "--core",
        action="store_true",
        help="show core options (by default is hidden unless -e ALL is passed)",
        dest="show_core",
    )
    register_env_select_flags(our, default=CliEnv())
    env_run_create_flags(our, mode="config")


def show_config(state: State) -> int:
    is_colored = state.conf.options.is_colored
    keys: list[str] = state.conf.options.list_keys_only
    is_first = True
    has_exception = False

    def _print_env(tox_env: ToxEnv) -> None:
        nonlocal is_first, has_exception
        if is_first:
            is_first = False
        else:
            print()  # noqa: T201
        print_section_header(is_colored, f"[testenv:{tox_env.conf.name}]")
        if not keys:
            print_key_value(is_colored, "type", type(tox_env).__name__)
        if print_conf(is_colored, tox_env.conf, keys):
            has_exception = True

    show_everything = state.conf.options.env.is_all
    done: set[str] = set()
    for name in state.envs.iter(package=True):  # now go through selected ones
        done.add(name)
        _print_env(state.envs[name])

    # environments may define core configuration flags, so we must exhaust first the environments to tell the core part
    if show_everything or state.conf.options.show_core:
        print()  # noqa: T201
        print_section_header(is_colored, "[tox]")
        if print_conf(is_colored, state.conf.core, keys):
            has_exception = True
    return -1 if has_exception else 0


def _colored(is_colored: bool, color: int, msg: str) -> str:  # noqa: FBT001
    return f"{color}{msg}{Fore.RESET}" if is_colored else msg


def print_section_header(is_colored: bool, name: str) -> None:  # noqa: FBT001
    print(_colored(is_colored, Fore.YELLOW, name))  # noqa: T201


def print_comment(is_colored: bool, comment: str) -> None:  # noqa: FBT001
    print(_colored(is_colored, Fore.CYAN, comment))  # noqa: T201


def print_key_value(is_colored: bool, key: str, value: str, multi_line: bool = False) -> None:  # noqa: FBT001, FBT002
    print(_colored(is_colored, Fore.GREEN, key), end="")  # noqa: T201
    print(" =", end="")  # noqa: T201
    if multi_line:
        print()  # noqa: T201
        value_str = indent(value, prefix="  ")
    else:
        print(" ", end="")  # noqa: T201
        value_str = value
    print(value_str)  # noqa: T201


def print_conf(is_colored: bool, conf: ConfigSet, keys: Iterable[str]) -> bool:  # noqa: FBT001
    has_exception = False
    for key in keys or conf:
        if key not in conf:
            continue
        key = conf.primary_key(key)  # noqa: PLW2901
        try:
            value = conf[key]
            as_str, multi_line = stringify(value)
        except Exception as exception:  # because e.g. the interpreter cannot be found
            if os.environ.get("_TOX_SHOW_CONFIG_RAISE"):  # pragma: no branch
                raise  # pragma: no cover
            as_str, multi_line = _colored(is_colored, Fore.LIGHTRED_EX, f"# Exception: {exception!r}"), False
            has_exception = True
        if multi_line and "\n" not in as_str:
            multi_line = False
        print_key_value(is_colored, key, as_str, multi_line=multi_line)
    unused = conf.unused()
    if unused and not keys:
        print_comment(is_colored, f"# !!! unused: {', '.join(unused)}")
    return has_exception



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from tox.pytest import ToxProjectCreator


def out_no_src(path: Path) -> str:
    return (
        f"ROOT: No loadable tox.ini or setup.cfg or pyproject.toml or tox.toml found, assuming empty tox.ini at {path}"
        f"\ndefault environments:\npy -> [no description]\n"
    )


def test_no_src_cwd(tox_project: ToxProjectCreator) -> None:
    project = tox_project({})
    outcome = project.run("l")
    outcome.assert_success()
    assert outcome.out == out_no_src(project.path)
    assert outcome.state.conf.src_path == (project.path / "tox.ini")


def test_no_src_has_py_project_toml_above(tox_project: ToxProjectCreator, tmp_path: Path) -> None:
    (tmp_path / "pyproject.toml").write_text("")
    project = tox_project({})
    outcome = project.run("l")
    outcome.assert_success()
    assert outcome.out == out_no_src(tmp_path)
    assert outcome.state.conf.src_path == (tmp_path / "tox.ini")


def test_no_src_root_dir(tox_project: ToxProjectCreator, tmp_path: Path) -> None:
    root = tmp_path / "root"
    root.mkdir()
    project = tox_project({})
    outcome = project.run("l", "--root", str(root))
    outcome.assert_success()
    assert outcome.out == out_no_src(root)
    assert outcome.state.conf.src_path == (root / "tox.ini")


def test_bad_src_content(tox_project: ToxProjectCreator, tmp_path: Path) -> None:
    project = tox_project({})

    outcome = project.run("l", "-c", str(tmp_path / "setup.cfg"))
    outcome.assert_failed()
    assert outcome.out == f"ROOT: HandledError| config file {tmp_path / 'setup.cfg'} does not exist\n"


def test_malformed_config_does_not_prevent_help(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.toml": "deps =\n    mypy\n"})
    outcome = project.run("--help")
    outcome.assert_success()
    assert "usage: tox" in outcome.out


def test_malformed_toml_in_dir_reports_error(tox_project: ToxProjectCreator) -> None:
    """Config discovery in a directory should report TOML parse errors instead of silently ignoring them."""
    project = tox_project({})
    # Write a pyproject.toml with an invalid TOML escape sequence (unescaped backslash)
    (project.path / "pyproject.toml").write_text('[tool.tox]\ntest = "c:\\path"\n', encoding="utf-8")
    outcome = project.run("l", "-c", str(project.path))
    outcome.assert_failed()
    assert "failed loading" in outcome.out


def test_toml_native_preferred_over_legacy_tox_ini(tox_project: ToxProjectCreator) -> None:
    """When pyproject.toml has both legacy_tox_ini and native TOML config, native TOML should win."""
    pyproject = """\
[tool.tox]
legacy_tox_ini = \"\"\"
[tox]
min_version = 4.21
[testenv]
commands = python -c "print('legacy')"
\"\"\"
env_list = ["native"]

[tool.tox.env_run_base]
package = "skip"
commands = [["python", "-c", "print('native')"]]
"""
    project = tox_project({"pyproject.toml": pyproject})
    outcome = project.run("l")
    outcome.assert_success()
    assert "native" in outcome.out
    assert "legacy" not in outcome.out



"""Provides configuration values from tox.ini files."""

from __future__ import annotations

import logging
import os
from configparser import ConfigParser
from pathlib import Path
from typing import Any, ClassVar

from platformdirs import user_config_dir

from tox.config.loader.api import ConfigLoadArgs
from tox.config.loader.ini import IniLoader
from tox.config.source.ini_section import CORE

DEFAULT_CONFIG_FILE = Path(user_config_dir("tox")) / "config.ini"


class IniConfig:
    TOX_CONFIG_FILE_ENV_VAR = "TOX_USER_CONFIG_FILE"
    STATE: ClassVar[dict[bool | None, str]] = {None: "failed to parse", True: "active", False: "missing"}

    def __init__(self) -> None:
        config_file = os.environ.get(self.TOX_CONFIG_FILE_ENV_VAR, None)
        self.is_env_var = config_file is not None
        self.config_file = Path(config_file if config_file is not None else DEFAULT_CONFIG_FILE)
        self._cache: dict[tuple[str, type[Any]], Any] = {}
        self.has_config_file: bool | None = self.config_file.exists()
        self.ini: IniLoader | None = None

        if self.has_config_file:
            self.config_file = self.config_file.absolute()
            try:
                parser = ConfigParser(interpolation=None)
                with self.config_file.open() as file_handler:
                    parser.read_file(file_handler)
                self.has_tox_section = parser.has_section(CORE.key)
                if self.has_tox_section:
                    self.ini = IniLoader(CORE, parser, overrides=[], core_section=CORE)
            except Exception as exception:  # noqa: BLE001
                logging.error("failed to read config file %s because %r", self.config_file, exception)  # noqa: TRY400
                self.has_config_file = None

    def get(self, key: str, of_type: type[Any]) -> Any:
        cache_key = key, of_type
        if cache_key in self._cache:
            result = self._cache[cache_key]
        else:
            try:
                if self.ini is None:  # pragma: no cover # this can only happen if we don't call __bool__ firsts
                    result = None
                else:
                    source = "file"
                    args = ConfigLoadArgs(chain=[key], name=CORE.prefix, env_name=None)
                    value = self.ini.load(key, of_type=of_type, conf=None, factory=None, args=args)
                    result = value, source
            except KeyError:  # just not found
                result = None
            except Exception as exception:  # noqa: BLE001
                logging.warning("%s key %s as type %r failed with %r", self.config_file, key, of_type, exception)
                result = None
        self._cache[cache_key] = result
        return result

    def __bool__(self) -> bool:
        return bool(self.has_config_file) and bool(self.has_tox_section)

    @property
    def epilog(self) -> str:
        # text to show within the parsers epilog
        return (
            f"{os.linesep}config file {str(self.config_file)!r} {self.STATE[self.has_config_file]} "
            f"(change{'d' if self.is_env_var else ''} via env var {self.TOX_CONFIG_FILE_ENV_VAR})"
        )



from __future__ import annotations

import io
import json
import tarfile
from textwrap import dedent
from typing import TYPE_CHECKING, Any

import pytest

from tox.execute.local_sub_process import LocalSubprocessExecuteStatus
from tox.tox_env.python.virtual_env.package.pyproject import Pep517VirtualEnvFrontend

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture

    from tox.pytest import ToxProjectCreator


@pytest.mark.parametrize(
    "pkg_type",
    ["editable-legacy", "editable", "sdist", "sdist-wheel", "wheel"],
)
def test_tox_ini_package_type_valid(tox_project: ToxProjectCreator, pkg_type: str) -> None:
    proj = tox_project({"tox.ini": f"[testenv]\npackage={pkg_type}", "pyproject.toml": ""})
    result = proj.run("c", "-k", "package_tox_env_type")
    result.assert_success()
    res = result.env_conf("py")["package"]
    assert res == pkg_type
    got_type = result.env_conf("py")["package_tox_env_type"]
    assert got_type == "virtualenv-pep-517"


def test_tox_ini_package_type_invalid(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=bad", "pyproject.toml": ""})
    result = proj.run("c", "-k", "package_tox_env_type")
    result.assert_failed()
    msg = (
        " invalid package config type bad requested,"
        " must be one of wheel, sdist, sdist-wheel, editable, editable-legacy, deps-only, skip"
    )
    assert msg in result.out


def test_get_package_deps_different_extras(pkg_with_extras_project: Path, tox_project: ToxProjectCreator) -> None:
    ini = "[testenv:a]\npackage=editable-legacy\nextras=docs\n[testenv:b]\npackage=sdist\nextras=format"
    proj = tox_project({"tox.ini": ini})
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--root", str(pkg_with_extras_project), "-e", "a,b")
    result.assert_success()
    installs = {
        i[0][0].conf.name: i[0][3].cmd[5:]
        for i in execute_calls.call_args_list
        if i[0][3].run_id.startswith("install_package_deps")
    }
    assert installs == {
        "a": [
            "colorama>=0.4.6",
            "platformdirs>=4.3.8",
            "setuptools",
            "sphinx-rtd-theme<1,>=0.4.3",
            "sphinx>=3",
        ],
        "b": ["black>=3", "colorama>=0.4.6", "flake8", "platformdirs>=4.3.8"],
    }


def test_package_root_via_root(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    ini = f"[tox]\npackage_root={demo_pkg_inline}\n[testenv]\npackage=wheel\nwheel_build_env=.pkg"
    proj = tox_project({"tox.ini": ini, "pyproject.toml": ""})
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()


def test_package_root_via_testenv(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    ini = f"[testenv]\npackage=wheel\nwheel_build_env=.pkg\npackage_root={demo_pkg_inline}"
    proj = tox_project({"tox.ini": ini, "pyproject.toml": ""})
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()


@pytest.mark.parametrize(
    ("conf", "extra", "deps"),
    [
        pytest.param("[project]", "", [], id="no_deps"),
        pytest.param("[project]\ndependencies=['B', 'A']", "", ["A", "B"], id="deps"),
        pytest.param(
            "[project]\ndependencies=['A']\noptional-dependencies.alpha=['B']\noptional-dependencies.beta=['C']",
            "alpha",
            ["A", "B"],
            id="deps_with_one_extra",
        ),
        pytest.param(
            "[project]\ndependencies=['A']\noptional-dependencies.alpha=['B']\noptional-dependencies.beta=['C']",
            "alpha,beta",
            ["A", "B", "C"],
            id="deps_with_two_extra",
        ),
        pytest.param(
            "[project]\ndependencies=['A']\noptional-dependencies.alpha=[]",
            "alpha",
            ["A"],
            id="deps_with_empty_extra",
        ),
        pytest.param(
            "[project]\ndependencies=['A']\ndynamic=['optional-dependencies']",
            "",
            ["A"],
            id="deps_with_dynamic_optional_no_extra",
        ),
        pytest.param(
            dedent(
                """
                [project]
                name='foo'
                dependencies=['foo[alpha]']
                optional-dependencies.alpha=['A']""",
            ),
            "",
            ["A"],
            id="deps_reference_extra",
        ),
        pytest.param(
            dedent(
                """
                [project]
                name='foo'
                dependencies=['A']
                optional-dependencies.alpha=['B']
                optional-dependencies.beta=['foo[alpha]']""",
            ),
            "beta",
            ["A", "B"],
            id="deps_with_recursive_extra",
        ),
        pytest.param(
            dedent(
                """
                [project]
                name='foo'
                dependencies=['A']
                optional-dependencies.alpha=['B']
                optional-dependencies.beta=['foo[alpha]']
                optional-dependencies.delta=['foo[beta]', 'D']""",
            ),
            "delta",
            ["A", "B", "D"],
            id="deps_with_two_recursive_extra",
        ),
        pytest.param(
            dedent(
                """
                [project]
                name='foo'
                optional-dependencies.alpha=['foo[beta]', 'A']
                optional-dependencies.beta=['foo[alpha]', 'B']""",
            ),
            "alpha",
            ["A", "B"],
            id="deps_with_circular_recursive_extra",
        ),
    ],
)
def test_pyproject_deps_from_static(
    tox_project: ToxProjectCreator,
    demo_pkg_inline: Path,
    conf: str,
    extra: str,
    deps: list[str],
) -> None:
    toml = f"{(demo_pkg_inline / 'pyproject.toml').read_text()}{conf}"
    proj = tox_project({"tox.ini": f"[testenv]\nextras={extra}", "pyproject.toml": toml}, base=demo_pkg_inline)
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()

    expected_calls = [(".pkg", "_optional_hooks"), (".pkg", "get_requires_for_build_sdist"), (".pkg", "build_sdist")]
    if deps:
        expected_calls.append(("py", "install_package_deps"))
    expected_calls.extend((("py", "install_package"), (".pkg", "_exit")))
    found_calls = [(i[0][0].conf.name, i[0][3].run_id) for i in execute_calls.call_args_list]
    assert found_calls == expected_calls

    if deps:
        expected_args = ["python", "-I", "-m", "pip", "install", *deps]
        args = execute_calls.call_args_list[-3][0][3].cmd
        assert expected_args == args


def test_pyproject_invalid_extra_static(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    conf = "[project]\nname='demo'\ndependencies=['A']\noptional-dependencies.alpha=['B']"
    toml = f"{(demo_pkg_inline / 'pyproject.toml').read_text()}{conf}"
    proj = tox_project({"tox.ini": "[testenv]\nextras=typo", "pyproject.toml": toml}, base=demo_pkg_inline)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_failed()
    assert "extras not found for package demo: typo (available: alpha)" in result.out


def test_pyproject_invalid_extra_no_optional_deps(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    conf = "[project]\nname='demo'"
    toml = f"{(demo_pkg_inline / 'pyproject.toml').read_text()}{conf}"
    proj = tox_project({"tox.ini": "[testenv]\nextras=alpha", "pyproject.toml": toml}, base=demo_pkg_inline)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_failed()
    assert "extras not found for package demo: alpha (available: none)" in result.out


@pytest.mark.parametrize(
    ("metadata", "dynamic", "extra", "deps"),
    [
        pytest.param("Requires-Dist: A", "['dependencies']", "", ["A"], id="deps"),
        pytest.param(
            "Provides-Extra: alpha\nRequires-Dist: A\nRequires-Dist: B;extra=='alpha'",
            "['dependencies']",
            "alpha",
            ["A", "B"],
            id="deps_extra",
        ),
        pytest.param(
            "Provides-Extra: alpha\nRequires-Dist: A\nRequires-Dist: B;extra=='alpha'",
            "['optional-dependencies']",
            "alpha",
            ["A", "B"],
            id="deps_extra_dynamic_opt",
        ),
    ],
)
def test_pyproject_deps_static_with_dynamic(  # noqa: PLR0913
    tox_project: ToxProjectCreator,
    demo_pkg_inline: Path,
    monkeypatch: pytest.MonkeyPatch,
    metadata: str,
    dynamic: str,
    extra: str,
    deps: list[str],
) -> None:
    monkeypatch.setenv("METADATA_EXTRA", metadata)
    toml = f"{(demo_pkg_inline / 'pyproject.toml').read_text()}[project]\ndynamic={dynamic}"
    ini = f"[testenv]\nextras={extra}\n[testenv:.pkg]\npass_env=METADATA_EXTRA"
    proj = tox_project({"tox.ini": ini, "pyproject.toml": toml}, base=demo_pkg_inline)

    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()

    expected_calls = [
        (".pkg", "_optional_hooks"),
        (".pkg", "get_requires_for_build_sdist"),
        (".pkg", "get_requires_for_build_wheel"),
        (".pkg", "build_wheel"),
        (".pkg", "build_sdist"),
        ("py", "install_package_deps"),
        ("py", "install_package"),
        (".pkg", "_exit"),
    ]
    found_calls = [(i[0][0].conf.name, i[0][3].run_id) for i in execute_calls.call_args_list]
    assert found_calls == expected_calls

    args = execute_calls.call_args_list[-3][0][3].cmd
    assert args == ["python", "-I", "-m", "pip", "install", *deps]


def test_pyproject_no_build_editable_fallback(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    proj = tox_project({"tox.ini": "[tox]\nenv_list=a,b"}, base=demo_pkg_inline)
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "-e", "a,b", "--notest", "--develop")
    result.assert_success()
    warning = (
        ".pkg: package config for a, b is editable, however the build backend build does not support PEP-660, "
        "falling back to editable-legacy - change your configuration to it"
    )
    assert warning in result.out.splitlines()

    expected_calls = [
        (".pkg", "_optional_hooks"),
        (".pkg", "get_requires_for_build_wheel"),
        (".pkg", "build_wheel"),
        (".pkg", "get_requires_for_build_sdist"),
        ("a", "install_package"),
        ("b", "install_package"),
        (".pkg", "_exit"),
    ]
    found_calls = [(i[0][0].conf.name, i[0][3].run_id) for i in execute_calls.call_args_list]
    assert found_calls == expected_calls


@pytest.mark.parametrize("package", ["sdist", "sdist-wheel", "wheel", "editable"])
def test_pep517_pkg_env_rejects_deps(tox_project: ToxProjectCreator, demo_pkg_setuptools: Path, package: str) -> None:
    ini = f"[testenv]\npackage={package}\n[pkgenv]\ndeps = A"
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_setuptools)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_failed()
    assert "does not support the deps configuration" in result.out


def test_pep517_pkg_env_allows_deps_for_editable_legacy(
    tox_project: ToxProjectCreator,
    demo_pkg_setuptools: Path,
) -> None:
    ini = "[testenv]\npackage=editable-legacy\n[pkgenv]\ndeps = A"
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_setuptools)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()
    assert "does not support the deps configuration" not in result.out


def test_pep517_pkg_env_rejects_deps_via_testenv(
    tox_project: ToxProjectCreator,
    demo_pkg_setuptools: Path,
) -> None:
    ini = "[testenv]\npackage=sdist\n[testenv:.pkg]\ndeps = B"
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_setuptools)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_failed()
    assert "does not support the deps configuration" in result.out


def test_pyproject_build_editable_and_wheel(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    # test that build wheel and build editable are cached separately

    ini = """
    [testenv:.pkg]
    set_env= BACKEND_HAS_EDITABLE=1
    [testenv:a,b]
    package = editable
    [testenv:c,d]
    package = wheel
    """
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_inline)
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)

    result = proj.run("r", "-e", "a,b,c,d", "--notest", "--workdir", str(proj.path / ".tox"))

    result.assert_success()
    found_calls = [(i[0][0].conf.name, i[0][3].run_id) for i in execute_calls.call_args_list]
    assert found_calls == [
        (".pkg", "_optional_hooks"),
        (".pkg", "get_requires_for_build_wheel"),
        (".pkg", "build_editable"),
        ("a", "install_package"),
        ("b", "install_package"),
        (".pkg", "build_wheel"),
        ("c", "install_package"),
        ("d", "install_package"),
        (".pkg", "_exit"),
    ]


def test_pyproject_config_settings_sdist(
    tox_project: ToxProjectCreator,
    demo_pkg_setuptools: Path,
    mocker: MockerFixture,
) -> None:
    ini = """
    [tox]
    env_list = sdist

    [testenv]
    wheel_build_env = .pkg
    package = sdist

    [testenv:.pkg]
    config_settings_get_requires_for_build_sdist = A = 1
    config_settings_build_sdist = B = 2
    config_settings_get_requires_for_build_wheel = C = 3
    config_settings_prepare_metadata_for_build_wheel = D = 4
    """
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_setuptools)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)

    write_stdin = mocker.spy(LocalSubprocessExecuteStatus, "write_stdin")

    result = proj.run("r", "--notest", from_cwd=proj.path)
    result.assert_success()

    found = {
        message["cmd"]: message["kwargs"]["config_settings"]
        for message in [json.loads(call[0][1]) for call in write_stdin.call_args_list]
        if not message["cmd"].startswith("_")
    }
    assert found == {
        "build_sdist": {"B": "2"},
        "get_requires_for_build_sdist": {"A": "1"},
        "get_requires_for_build_wheel": {"C": "3"},
        "prepare_metadata_for_build_wheel": {"D": "4"},
    }


def test_pyproject_config_settings_wheel(
    tox_project: ToxProjectCreator,
    demo_pkg_setuptools: Path,
    mocker: MockerFixture,
) -> None:
    ini = """
    [tox]
    env_list = wheel

    [testenv]
    wheel_build_env = .pkg
    package = wheel

    [testenv:.pkg]
    config_settings_get_requires_for_build_wheel = C = 3
    config_settings_prepare_metadata_for_build_wheel = D = 4
    config_settings_build_wheel = E = 5
    """
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_setuptools)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)

    write_stdin = mocker.spy(LocalSubprocessExecuteStatus, "write_stdin")
    mocker.patch.object(Pep517VirtualEnvFrontend, "_can_skip_prepare", return_value=False)

    result = proj.run("r", "--notest", from_cwd=proj.path)
    result.assert_success()

    found = {
        message["cmd"]: message["kwargs"]["config_settings"]
        for message in [json.loads(call[0][1]) for call in write_stdin.call_args_list]
        if not message["cmd"].startswith("_")
    }
    assert found == {
        "get_requires_for_build_wheel": {"C": "3"},
        "prepare_metadata_for_build_wheel": {"D": "4"},
        "build_wheel": {"E": "5"},
    }


def test_pyproject_config_settings_editable(
    tox_project: ToxProjectCreator,
    demo_pkg_setuptools: Path,
    mocker: MockerFixture,
) -> None:
    ini = """
    [tox]
    env_list = editable

    [testenv:.pkg]
    config_settings_get_requires_for_build_editable = F = 6
    config_settings_prepare_metadata_for_build_editable = G = 7
    config_settings_build_editable = H = 8

    [testenv]
    wheel_build_env = .pkg
    package = editable
    """
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_setuptools)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)

    write_stdin = mocker.spy(LocalSubprocessExecuteStatus, "write_stdin")
    mocker.patch.object(Pep517VirtualEnvFrontend, "_can_skip_prepare", return_value=False)

    result = proj.run("r", "--notest", from_cwd=proj.path)
    result.assert_success()

    found = {
        message["cmd"]: message["kwargs"]["config_settings"]
        for message in [json.loads(call[0][1]) for call in write_stdin.call_args_list]
        if not message["cmd"].startswith("_")
    }
    assert found == {
        "get_requires_for_build_editable": {"F": "6"},
        "prepare_metadata_for_build_editable": {"G": "7"},
        "build_editable": {"H": "8"},
    }


def test_pyproject_config_settings_editable_legacy(
    tox_project: ToxProjectCreator,
    demo_pkg_setuptools: Path,
    mocker: MockerFixture,
) -> None:
    ini = """
    [tox]
    env_list = editable

    [testenv:.pkg]
    config_settings_get_requires_for_build_sdist = A = 1
    config_settings_get_requires_for_build_wheel = C = 3
    config_settings_prepare_metadata_for_build_wheel = D = 4

    [testenv]
    wheel_build_env = .pkg
    package = editable-legacy
    """
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_setuptools)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)

    write_stdin = mocker.spy(LocalSubprocessExecuteStatus, "write_stdin")
    mocker.patch.object(Pep517VirtualEnvFrontend, "_can_skip_prepare", return_value=False)

    result = proj.run("r", "--notest", from_cwd=proj.path)
    result.assert_success()

    found = {
        message["cmd"]: message["kwargs"]["config_settings"]
        for message in [json.loads(call[0][1]) for call in write_stdin.call_args_list]
        if not message["cmd"].startswith("_")
    }
    assert found == {
        "get_requires_for_build_sdist": {"A": "1"},
        "get_requires_for_build_wheel": {"C": "3"},
        "prepare_metadata_for_build_wheel": {"D": "4"},
    }


def test_pyproject_config_settings_sdist_passed_to_pip_install(
    tox_project: ToxProjectCreator,
    demo_pkg_setuptools: Path,
    mocker: MockerFixture,
) -> None:
    toml = """
    [tox]
    env_list = ["sdist"]

    [env_run_base]
    wheel_build_env = ".pkg"
    package = "sdist"

    [env.".pkg"]
    config_settings_build_wheel = {X = "y", Z = "w"}
    """
    proj = tox_project({"tox.toml": dedent(toml)}, base=demo_pkg_setuptools)
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    mocker.spy(LocalSubprocessExecuteStatus, "write_stdin")

    result = proj.run("r", "--notest", from_cwd=proj.path)
    result.assert_success()

    pip_install_cmds = [i[0][3].cmd for i in execute_calls.call_args_list if i[0][3].run_id == "install_package"]
    assert pip_install_cmds, "expected at least one install_package call"
    cmd = pip_install_cmds[0]
    assert "--config-settings=X=y" in cmd
    assert "--config-settings=Z=w" in cmd


@pytest.mark.integration
@pytest.mark.usefixtures("enable_pip_pypi_access")
def test_pyproject_installpkg_pep517_envs(tox_project: ToxProjectCreator, pkg_with_pdm_backend: Path) -> None:
    # Regression test for #3512
    tox_ini = """
    [tox]
    envlist = dummy1,dummy2

    [testenv:dummy1]
    commands =
        python -c print(1)

    [testenv:dummy2]
    commands =
        python -c print(42)
    """
    sdist = pkg_with_pdm_backend / "dist" / "skeleton-0.1.1337.tar.gz"
    proj = tox_project({"tox.ini": tox_ini}, base=pkg_with_pdm_backend)
    result = proj.run("--installpkg", str(sdist))
    result.assert_success()


def test_sdist_wheel_package_type(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    ini = "[testenv]\npackage=sdist-wheel"
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_inline)
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()

    found_calls = [(i[0][0].conf.name, i[0][3].run_id) for i in execute_calls.call_args_list]

    # The .pkg env builds the sdist
    assert (".pkg", "build_sdist") in found_calls, f"expected .pkg to build_sdist, got {found_calls}"

    # A wheel is built from the extracted sdist (may be in .pkg or a wheel_build_env child)
    wheel_calls = [(i, env, rid) for i, (env, rid) in enumerate(found_calls) if rid == "build_wheel"]
    assert wheel_calls, f"expected build_wheel call, got {found_calls}"

    # The sdist-wheel flow: build_sdist must happen before the final build_wheel
    # (an earlier build_wheel may occur for metadata extraction)
    sdist_idx = found_calls.index((".pkg", "build_sdist"))
    last_wheel_idx = wheel_calls[-1][0]
    assert sdist_idx < last_wheel_idx, f"build_sdist must happen before final build_wheel, got {found_calls}"

    # The final package installed should be a wheel
    install_calls = [(env, rid) for env, rid in found_calls if rid == "install_package"]
    assert install_calls, "expected install_package call"


def test_sdist_wheel_config(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    ini = "[testenv]\npackage=sdist-wheel"
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_inline)
    result = proj.run("c", "-k", "package", "-k", "wheel_build_env")
    result.assert_success()
    res = result.env_conf("py")["package"]
    assert res == "sdist-wheel"
    # wheel_build_env should be present (reusing existing infrastructure)
    build_env = result.env_conf("py")["wheel_build_env"]
    assert build_env  # has a value (name depends on Python version)


def test_sdist_wheel_custom_build_env(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    ini = "[testenv]\npackage=sdist-wheel\nwheel_build_env=.custom-wheel-builder"
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_inline)
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()

    found_calls = [(i[0][0].conf.name, i[0][3].run_id) for i in execute_calls.call_args_list]
    env_names = {env for env, _ in found_calls}
    assert ".custom-wheel-builder" in env_names, f"expected custom child env name, got {env_names}"


def test_wheel_package_does_not_build_sdist(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    ini = "[testenv]\npackage=wheel"
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_inline)
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--notest")
    result.assert_success()
    # Verify that build_sdist was NOT called with plain wheel package type
    call_ids = [i[0][3].run_id for i in execute_calls.call_args_list]
    assert "build_sdist" not in call_ids
    assert "build_wheel" in call_ids


def test_sdist_wheel_rejects_path_traversal(
    tox_project: ToxProjectCreator, demo_pkg_inline: Path, mocker: MockerFixture
) -> None:
    ini = "[testenv]\npackage=sdist-wheel"
    proj = tox_project({"tox.ini": ini}, base=demo_pkg_inline)
    proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)

    original_build_sdist = Pep517VirtualEnvFrontend.build_sdist

    def _build_malicious_sdist(self: Pep517VirtualEnvFrontend, *args: Any, **kwargs: Any) -> object:
        result = original_build_sdist(self, *args, **kwargs)
        buf = io.BytesIO()
        with tarfile.open(fileobj=buf, mode="w:gz") as tar:
            info = tarfile.TarInfo(name="../escape.txt")
            info.size = 4
            tar.addfile(info, io.BytesIO(b"evil"))
        result.sdist.write_bytes(buf.getvalue())
        return result

    mocker.patch.object(Pep517VirtualEnvFrontend, "build_sdist", _build_malicious_sdist)
    result = proj.run("r", "--notest")
    result.assert_failed()
    assert "tar member '../escape.txt' would extract outside of" in result.out



"""Apply value substitution (replacement) on tox strings."""

from __future__ import annotations

import re
from configparser import SectionProxy
from functools import cache
from re import Pattern
from typing import TYPE_CHECKING

from tox.config.loader.replacer import ReplaceReference
from tox.config.loader.stringify import stringify

if TYPE_CHECKING:
    from collections.abc import Iterator

    from tox.config.loader.api import ConfigLoadArgs
    from tox.config.loader.ini import IniLoader
    from tox.config.main import Config
    from tox.config.sets import ConfigSet


class ReplaceReferenceIni(ReplaceReference):
    def __init__(self, conf: Config, loader: IniLoader) -> None:
        self.conf = conf
        self.loader = loader

    def __call__(self, value: str, conf_args: ConfigLoadArgs) -> str | None:  # noqa: C901
        # a return value of None indicates could not replace
        pattern = _replace_ref(self.loader.section.prefix or self.loader.section.name)
        match = pattern.match(value)
        if match:
            settings = match.groupdict()

            key = settings["key"]
            if settings["section"] is None and settings["full_env"]:
                settings["section"] = settings["full_env"]

            exception: Exception | None = None
            try:
                for src in self._config_value_sources(settings["env"], settings["section"], conf_args.env_name):
                    try:
                        if isinstance(src, SectionProxy):
                            return self._resolve_section_proxy(src, key, conf_args.env_name)
                        value = src.load(key, conf_args.chain)
                    except KeyError as exc:  # if fails, keep trying maybe another source can satisfy # noqa: PERF203
                        exception = exc
                    else:
                        as_str, _ = stringify(value)
                        return as_str.replace("#", r"\#")  # escape comment characters as these will be stripped
            except Exception as exc:  # noqa: BLE001
                exception = exc
            if exception is not None:
                if isinstance(exception, KeyError):  # if the lookup failed replace - else keep
                    default = settings["default"]
                    if default is not None:
                        return default
                    # we cannot raise here as that would mean users could not write factorials:
                    #   depends = {py39,py38}-{,b}
                else:
                    raise exception
        return None

    def _resolve_section_proxy(self, src: SectionProxy, key: str, env_name: str | None) -> str:
        """Resolve a key from a SectionProxy, returning empty string when factor filtering empties the value."""
        try:
            return self.loader.process_raw(self.conf, env_name, src[key])
        except KeyError:
            if key in src:
                # Key exists but factor filtering emptied the value.
                # For cross-section references this is a valid empty result,
                # not a missing key — the caller explicitly asked for this value.
                return ""
            raise

    def _config_value_sources(
        self, env: str | None, section: str | None, current_env: str | None
    ) -> Iterator[SectionProxy | ConfigSet]:
        if section is None:
            if env is not None and env in self.conf:
                yield self.conf.get_env(env)
            # if no section specified perhaps it's an unregistered config:
            # 1. try first from core conf
            yield self.conf.core
            # 2. and then fallback to our own environment
            if current_env is not None:
                yield self.conf.get_env(current_env)
            return

        # if there's a section, special handle the core section
        if section == self.loader.core_section.name:
            yield self.conf.core  # try via registered configs
        # prefer raw section values so substitutions resolve in the calling env's context
        if (proxy := self.loader.get_section(section)) is not None:
            yield proxy
        # fallback to the env's ConfigSet for registered-only keys
        if env is not None and env in self.conf:
            yield self.conf.get_env(env)


@cache
def _replace_ref(env: str | None) -> Pattern[str]:
    return re.compile(
        rf"""
    (\[(?P<full_env>{re.escape(env or ".*")}(:(?P<env>[^]]+))?|(?P<section>[-\w]+))])? # env/section
    (?P<key>[-a-zA-Z0-9_]+) # key
    (:(?P<default>.*))? # default value
    $
""",
        re.VERBOSE,
    )


__all__ = [
    "ReplaceReferenceIni",
]



from __future__ import annotations

import socket
import sys
from typing import Any

import pytest

from tox import __version__
from tox.journal.main import Journal


@pytest.fixture(scope="session")
def base_info() -> dict[str, Any]:
    return {
        "reportversion": "1",
        "toxversion": __version__,
        "platform": sys.platform,
        "host": socket.getfqdn(),
    }


def test_journal_enabled_default(base_info: dict[str, Any]) -> None:
    journal = Journal(enabled=True)
    assert bool(journal) is True
    assert journal.content == base_info


def test_journal_disabled_default() -> None:
    journal = Journal(enabled=False)
    assert bool(journal) is False
    assert journal.content == {}


def test_env_journal_enabled(base_info: dict[str, Any]) -> None:
    journal = Journal(enabled=True)
    env = journal.get_env_journal("a")
    assert journal.get_env_journal("a") is env
    env["demo"] = 1

    assert bool(env) is True
    base_info["testenvs"] = {"a": {"demo": 1}}
    assert journal.content == base_info


def test_env_journal_disabled() -> None:
    journal = Journal(enabled=False)
    env = journal.get_env_journal("a")
    assert bool(env) is False

    env["demo"] = 2
    assert journal.content == {"testenvs": {"a": {"demo": 2}}}



from __future__ import annotations

from typing import TYPE_CHECKING
from zipfile import ZipFile

import pytest

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from tox.pytest import ToxProjectCreator


@pytest.fixture(scope="session")
def pkg_with_extras_project_wheel(
    pkg_with_extras_project: Path,
    pkg_builder: Callable[[Path, Path, list[str], bool], Path],
) -> Path:
    dist = pkg_with_extras_project / "dist"
    pkg_builder(dist, pkg_with_extras_project, ["wheel"], False)
    return next(dist.iterdir())


def test_tox_install_pkg_wheel(tox_project: ToxProjectCreator, pkg_with_extras_project_wheel: Path) -> None:
    proj = tox_project({"tox.ini": "[testenv]\nextras=docs,format"})
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "-e", "py", "--installpkg", str(pkg_with_extras_project_wheel))
    result.assert_success()
    calls = [(i[0][0].conf.name, i[0][3].run_id, i[0][3].cmd[5:]) for i in execute_calls.call_args_list]
    deps = ["black>=3", "colorama>=0.4.6", "flake8", "platformdirs>=4.3.8", "sphinx-rtd-theme<1,>=0.4.3", "sphinx>=3"]
    expected = [
        ("py", "install_package_deps", deps),
        ("py", "install_package", ["--force-reinstall", "--no-deps", str(pkg_with_extras_project_wheel)]),
    ]
    assert calls == expected


@pytest.fixture(scope="session")
def pkg_with_sdist(
    pkg_with_extras_project: Path,
    pkg_builder: Callable[[Path, Path, list[str], bool], Path],
) -> Path:
    dist = pkg_with_extras_project / "dist"
    pkg_builder(dist, pkg_with_extras_project, ["sdist"], False)
    return next(dist.iterdir())


@pytest.fixture
def pkg_with_extras_project_sdist(
    pkg_with_extras_project: Path,
    pkg_builder: Callable[[Path, Path, list[str], bool], Path],
) -> Path:
    dist = pkg_with_extras_project / "sdist"
    pkg_builder(dist, pkg_with_extras_project, ["sdist"], False)
    return next(dist.iterdir())


def test_tox_install_pkg_sdist(tox_project: ToxProjectCreator, pkg_with_extras_project_sdist: Path) -> None:
    proj = tox_project({"tox.ini": "[testenv]\nextras=docs,format"})
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "-e", "py", "--installpkg", str(pkg_with_extras_project_sdist))
    result.assert_success()
    calls = [(i[0][0].conf.name, i[0][3].run_id, i[0][3].cmd[5:]) for i in execute_calls.call_args_list]
    deps = ["black>=3", "colorama>=0.4.6", "flake8", "platformdirs>=4.3.8", "sphinx-rtd-theme<1,>=0.4.3", "sphinx>=3"]
    assert calls == [
        (".pkg_external_sdist_meta", "install_requires", ["setuptools"]),
        (".pkg_external_sdist_meta", "_optional_hooks", []),
        (".pkg_external_sdist_meta", "get_requires_for_build_sdist", []),
        (".pkg_external_sdist_meta", "get_requires_for_build_wheel", []),  # required before prepare_metadata*
        (".pkg_external_sdist_meta", "prepare_metadata_for_build_wheel", []),
        ("py", "install_package_deps", deps),
        ("py", "install_package", ["--force-reinstall", "--no-deps", str(pkg_with_extras_project_sdist)]),
    ]


@pytest.mark.parametrize("mode", ["p", "le"])  # no need for r as is tested above
def test_install_pkg_via(tox_project: ToxProjectCreator, mode: str, pkg_with_extras_project_wheel: Path) -> None:
    proj = tox_project({"tox.ini": "[testenv]\npackage=wheel"})
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)

    result = proj.run(mode, "--installpkg", str(pkg_with_extras_project_wheel))

    result.assert_success()
    calls = [(i[0][0].conf.name, i[0][3].run_id) for i in execute_calls.call_args_list]
    assert calls == [("py", "install_package_deps"), ("py", "install_package")]


@pytest.mark.usefixtures("enable_pip_pypi_access")
def test_build_wheel_external_runs_once(
    tox_project: ToxProjectCreator, demo_pkg_inline: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    toml = """
    env_list = ["a", "b"]
    [env_run_base]
    package = "external"
    package_env = ".ext"
    commands = [["python", "-c", "from demo_pkg_inline import do; do()"]]
    [env.".ext"]
    deps = ["build"]
    package_glob = "{env_tmp_dir}/dist/*.whl"
    commands = [["pyproject-build", "-w", ".", "-o", "{env_tmp_dir}/dist"]]
    """
    project = tox_project({"tox.toml": toml})
    monkeypatch.delenv("PIP_EXTRA_INDEX_URL", raising=False)
    monkeypatch.delenv("PIP_INDEX_URL", raising=False)

    result = project.run("r", "--root", str(demo_pkg_inline), "--workdir", str(project.path))

    result.assert_success()
    assert result.out.count("pyproject-build") == 1
    assert result.out.count("greetings from demo_pkg_inline") == 2


@pytest.mark.usefixtures("enable_pip_pypi_access")
def test_build_wheel_external(
    tox_project: ToxProjectCreator, demo_pkg_inline: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    ini = """
    [testenv]
    package = external
    package_env = .ext
    commands =
        python -c 'from demo_pkg_inline import do; do()'

    [testenv:.ext]
    deps = build
    package_glob = {envtmpdir}{/}dist{/}*.whl
    commands =
        pyproject-build -w . -o {envtmpdir}{/}dist
    """
    project = tox_project({"tox.ini": ini})
    monkeypatch.delenv("PIP_EXTRA_INDEX_URL", raising=False)
    monkeypatch.delenv("PIP_INDEX_URL", raising=False)

    result = project.run("r", "--root", str(demo_pkg_inline), "--workdir", str(project.path))

    result.assert_success()
    assert "greetings from demo_pkg_inline" in result.out


def test_build_wheel_external_fail_build(tox_project: ToxProjectCreator) -> None:
    ini = """
    [testenv]
    package = external
    [testenv:.pkg_external]
    commands = xyz
    """
    project = tox_project({"tox.ini": ini})
    result = project.run("r")

    result.assert_failed()
    assert "stopping as failed to build package" in result.out, result.out


def test_build_wheel_external_fail_no_pkg(tox_project: ToxProjectCreator) -> None:
    ini = """
    [testenv]
    package = external
    """
    project = tox_project({"tox.ini": ini})
    result = project.run("r")

    result.assert_failed()
    assert "failed with no package found in " in result.out, result.out


def test_build_wheel_external_fail_many_pkg(tox_project: ToxProjectCreator) -> None:
    ini = """
    [testenv]
    package = external
    [testenv:.pkg_external]
    commands =
        python -c 'from pathlib import Path; (Path(r"{env_tmp_dir}") / "dist").mkdir()'
        python -c 'from pathlib import Path; (Path(r"{env_tmp_dir}") / "dist" / "a").write_text("")'
        python -c 'from pathlib import Path; (Path(r"{env_tmp_dir}") / "dist" / "b").write_text("")'
    """
    project = tox_project({"tox.ini": ini})
    result = project.run("r")

    result.assert_failed()
    assert "failed with found more than one package " in result.out, result.out


def test_tox_install_pkg_bad_wheel(tox_project: ToxProjectCreator, tmp_path: Path) -> None:
    wheel = tmp_path / "w.whl"
    with ZipFile(str(wheel), "w"):
        pass
    proj = tox_project({"tox.ini": "[testenv]"})
    result = proj.run("r", "-e", "py", "--installpkg", str(wheel))

    result.assert_failed()
    assert "failed with no .dist-info inside " in result.out, result.out


def test_tox_install_pkg_with_skip_install(
    tox_project: ToxProjectCreator,
    demo_pkg_inline: Path,
    demo_pkg_inline_wheel: Path,
) -> None:
    ini = "[testenv:foo]\nskip_install = true"
    project = tox_project({"tox.ini": ini, "pyproject.toml": (demo_pkg_inline / "pyproject.toml").read_text()})
    result = project.run("-e", "py", "--installpkg", str(demo_pkg_inline_wheel))
    result.assert_success()


def test_run_installpkg_targz(
    tox_project: ToxProjectCreator,
    pkg_with_sdist: Path,
    enable_pip_pypi_access: str | None,  # noqa: ARG001
) -> None:
    project = tox_project({
        "tox.ini": """
     [tox]
    envlist = base, flake8
    [testenv]
    package = sdist
    [testenv:base]
    [testenv:flake8]
    """
    })
    outcome = project.run(f"--installpkg={pkg_with_sdist}")
    outcome.assert_success()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, NoReturn, TypeVar

import pytest

from tox.config.loader.api import ConfigLoadArgs
from tox.config.loader.toml import TomlLoader
from tox.config.source.toml_pyproject import TomlPyProjectSection
from tox.config.types import Command, EnvList
from tox.report import HandledError

if TYPE_CHECKING:
    from types import UnionType


def test_toml_loader_load_raw() -> None:
    loader = TomlLoader(TomlPyProjectSection.from_key("tox.env.A"), [], {"a": 1, "c": False}, {}, set())
    assert loader.load_raw("a", None, "A") == 1


def test_toml_loader_load_repr() -> None:
    loader = TomlLoader(TomlPyProjectSection.from_key("tox.env.A"), [], {"a": 1}, {}, set())
    assert repr(loader) == "TomlLoader(env.A, {'a': 1})"


def test_toml_loader_found_keys() -> None:
    loader = TomlLoader(TomlPyProjectSection.from_key("tox.env.A"), [], {"a": 1, "c": False}, {}, set())
    assert loader.found_keys() == {"a", "c"}


def factory_na(obj: object) -> NoReturn:
    raise NotImplementedError


V = TypeVar("V")


def perform_load(value: Any, of_type: type[V] | UnionType) -> V:
    env_name, key = "A", "k"
    loader = TomlLoader(TomlPyProjectSection.from_key(f"tox.env.{env_name}"), [], {key: value}, {}, set())
    args = ConfigLoadArgs(None, env_name, env_name)
    return loader.load(key, of_type, factory_na, None, args)


_PREFIX = r"failed to load A\.k: "


def test_toml_loader_str_ok() -> None:
    assert perform_load("s", str) == "s"


def test_toml_loader_str_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"1 is not of type 'str'"):
        perform_load(1, str)


def test_toml_loader_bool_ok() -> None:
    assert perform_load(True, bool) is True


def test_toml_loader_bool_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"'true' is not of type 'bool'"):
        perform_load("true", bool)


def test_toml_loader_list_ok() -> None:
    assert perform_load(["a"], list[str]) == ["a"]


def test_toml_loader_list_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"\{\} is not list"):
        perform_load({}, list[str])


def test_toml_loader_list_nok_element() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"2 is not of type 'str'"):
        perform_load(["a", 2], list[str])


def test_toml_loader_dict_ok() -> None:
    assert perform_load({"a": "1"}, dict[str, str]) == {"a": "1"}


def test_toml_loader_dict_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"\{'a'\} is not dictionary"):
        perform_load({"a"}, dict[str, str])


def test_toml_loader_dict_nok_key() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"1 is not of type 'str'"):
        perform_load({"a": 1, 1: "2"}, dict[str, int])


def test_toml_loader_dict_nok_value() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"'2' is not of type 'int'"):
        perform_load({"a": 1, "b": "2"}, dict[str, int])


def test_toml_loader_path_ok() -> None:
    assert perform_load("/w", Path) == Path("/w")


def test_toml_loader_path_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"1 is not of type 'str'"):
        perform_load(1, Path)


def test_toml_loader_command_ok() -> None:
    commands = perform_load([["a", "b"], ["c"]], list[Command])
    assert isinstance(commands, list)
    assert len(commands) == 2
    assert all(isinstance(i, Command) for i in commands)

    assert commands[0].args == ["a", "b"]
    assert commands[1].args == ["c"]


def test_toml_loader_command_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"1 is not of type 'str'"):
        perform_load([["a", 1]], list[Command])


def test_toml_loader_env_list_ok() -> None:
    res = perform_load(["a", "b"], EnvList)
    assert isinstance(res, EnvList)
    assert list(res) == ["a", "b"]


def test_toml_loader_env_list_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"env_list items must be strings or product dicts, got int"):
        perform_load(["a", 1], EnvList)


def test_toml_loader_list_optional_ok() -> None:
    assert perform_load(["a", None], list[str | None]) == ["a", None]


def test_toml_loader_list_optional_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"1 is not union of str, NoneType"):
        perform_load(["a", None, 1], list[str | None])


def test_toml_loader_list_literal_ok() -> None:
    assert perform_load(["a", "b"], list[Literal["a", "b"]]) == ["a", "b"]


def test_toml_loader_list_literal_nok() -> None:
    with pytest.raises(HandledError, match=_PREFIX + r"'c' is not one of literal 'a','b'"):
        perform_load(["a", "c"], list[Literal["a", "b"]])


def test_toml_loader_union_list_or_str_with_list() -> None:
    assert perform_load(["a", "b"], list[str] | str) == ["a", "b"]


def test_toml_loader_union_list_or_str_with_str() -> None:
    assert perform_load("a", list[str] | str) == "a"



from __future__ import annotations

import os
import platform
import sys
from configparser import ConfigParser
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from tox.config.types import Command
from tox.execute.request import shell_cmd

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from pytest_mock import MockerFixture

    from tox.pytest import MonkeyPatch, ToxProjectCreator


def test_show_config_default_run_env(tox_project: ToxProjectCreator, monkeypatch: MonkeyPatch) -> None:
    py_ver = sys.version_info[0:2]
    name = f"py{py_ver[0]}{py_ver[1]}" if platform.python_implementation() == "CPython" else "pypy3"
    project = tox_project({"tox.ini": f"[tox]\nenv_list = {name}\n[testenv:{name}]\ncommands={{posargs}}"})
    result = project.run("c", "-e", name, "--core", "--", "magic")
    state = result.state
    assert state.args == ("c", "-e", name, "--core", "--", "magic")
    outcome = list(state.envs.iter(only_active=False))
    assert outcome == [name]
    monkeypatch.delenv("TERM", raising=False)  # disable conditionally set flag
    parser = ConfigParser(interpolation=None)
    parser.read_string(result.out)
    assert list(parser.sections()) == [f"testenv:{name}", "tox"]


def test_show_config_commands(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.ini": """
        [tox]
        env_list = py
        no_package = true
        [testenv]
        commands_pre =
            python -c 'import sys; print("start", sys.executable)'
        commands =
            pip config list
            pip list
        commands_post =
            python -c 'import sys; print("end", sys.executable)'
        """,
        },
    )
    outcome = project.run("c")
    outcome.assert_success()
    env_config = outcome.env_conf("py")
    assert env_config["commands_pre"] == [Command(args=["python", "-c", 'import sys; print("start", sys.executable)'])]
    assert env_config["commands"] == [
        Command(args=["pip", "config", "list"]),
        Command(args=["pip", "list"]),
    ]
    assert env_config["commands_post"] == [Command(args=["python", "-c", 'import sys; print("end", sys.executable)'])]


def test_show_config_filter_keys(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv]\nmagic=yes"})
    outcome = project.run("c", "-e", "py", "-k", "no_package", "env_name", "--core")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:py]\nenv_name = py\n\n[tox]\nno_package = False\n", "")


def test_show_config_unused(tox_project: ToxProjectCreator) -> None:
    tox_ini = "[testenv]\nok=false\n[testenv:py]\nmagical=yes\nmagic=yes"
    outcome = tox_project({"tox.ini": tox_ini}).run("c", "-e", "py")
    outcome.assert_success()
    assert "\n# !!! unused: magic, magical\n" in outcome.out


def test_show_config_py_ver_impl_constants(tox_project: ToxProjectCreator) -> None:
    tox_ini = "[testenv]\npackage=skip\ndeps= {py_impl}{py_dot_ver}"
    outcome = tox_project({"tox.ini": tox_ini}).run("c", "-e", "py", "-k", "py_dot_ver", "py_impl", "deps")
    outcome.assert_success()
    py_ver = ".".join(str(i) for i in sys.version_info[0:2])
    impl = sys.implementation.name
    assert outcome.out == f"[testenv:py]\npy_dot_ver = {py_ver}\npy_impl = {impl}\ndeps = {impl}{py_ver}\n"


def test_show_config_exception(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.ini": """
        [testenv:a]
        base_python = missing-python
        """,
        },
    )
    outcome = project.run("c", "-e", "a", "-k", "env_site_packages_dir", raise_on_config_fail=False)
    outcome.assert_failed(code=-1)
    txt = (
        "\nenv_site_packages_dir = # Exception: "
        "Skip(\"failed to find interpreter for Builtin discover of python_spec='missing-python'"
    )
    assert txt in outcome.out


def test_show_config_empty_install_command_exception(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv:a]\ninstall_command="})
    outcome = project.run("c", "-e", "a", "-k", "install_command", raise_on_config_fail=False)
    outcome.assert_failed(code=-1)
    txt = "\ninstall_command = # Exception: ValueError(\"attempting to parse '' into a command failed\")"
    assert txt in outcome.out


@pytest.mark.parametrize(
    ("filename", "content"),
    [
        pytest.param(
            "tox.toml",
            """
            [env_run_base]
            package = "skip"
            install_command = ["echo", "CUSTOM", "{packages}"]
            commands = [["python", "-c", "pass"]]
            """,
            id="tox.toml",
        ),
        pytest.param(
            "pyproject.toml",
            """
            [tool.tox]
            env_list = ["py"]
            [tool.tox.env_run_base]
            package = "skip"
            install_command = ["echo", "CUSTOM", "{packages}"]
            commands = [["python", "-c", "pass"]]
            """,
            id="pyproject.toml",
        ),
    ],
)
def test_show_config_install_command_toml(tox_project: ToxProjectCreator, filename: str, content: str) -> None:
    project = tox_project({filename: content})
    outcome = project.run("c", "-k", "install_command")
    outcome.assert_success()
    expected_cmd = shell_cmd(["echo", "CUSTOM", "{packages}"])
    assert f"install_command = {expected_cmd}" in outcome.out


def test_show_config_invalid_python_exit_code(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.ini": """
        [testenv:a]
        base_python = py47
        """,
        },
    )
    outcome = project.run("c", "-e", "a", raise_on_config_fail=False)
    outcome.assert_failed(code=-1)
    assert "# Exception:" in outcome.out


def test_show_config_valid_env_exit_code_zero(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.ini": """
        [tox]
        no_package = true
        [testenv:a]
        commands = python -c 'pass'
        """,
        },
    )
    outcome = project.run("c", "-e", "a")
    outcome.assert_success()


@pytest.mark.parametrize("stdout_is_atty", [True, False])
def test_pass_env_config_default(tox_project: ToxProjectCreator, stdout_is_atty: bool, mocker: MockerFixture) -> None:
    mocker.patch("sys.stdout.isatty", return_value=stdout_is_atty)
    project = tox_project({"tox.ini": ""})
    outcome = project.run("c", "-e", "py", "-k", "pass_env")
    pass_env = outcome.env_conf("py")["pass_env"]
    is_win = sys.platform == "win32"
    expected = (
        []
        + (["APPDATA"] if is_win else [])
        + ["CC", "CCSHARED", "CFLAGS"]
        + (["COMSPEC"] if is_win else [])
        + ["CPPFLAGS", "CURL_CA_BUNDLE", "CXX", "FORCE_COLOR", "HOME", "LANG"]
        + ["LANGUAGE", "LDFLAGS", "LD_LIBRARY_PATH"]
        + (["LOCALAPPDATA"] if is_win else [])
        + (["MSYSTEM"] if is_win else [])
        + ["NETRC"]
        + (["NIX_LD", "NIX_LD_LIBRARY_PATH"] if not is_win else [])
        + ["NO_COLOR"]
        + (["NUMBER_OF_PROCESSORS", "PATHEXT"] if is_win else [])
        + ["PIP_*", "PKG_CONFIG", "PKG_CONFIG_PATH", "PKG_CONFIG_SYSROOT_DIR"]
        + (["PROCESSOR_ARCHITECTURE"] if is_win else [])
        + (["PROGRAMDATA"] if is_win else [])
        + (["PROGRAMFILES"] if is_win else [])
        + (["PROGRAMFILES(x86)"] if is_win else [])
        + ["PYTHON_GIL", "REQUESTS_CA_BUNDLE"]
        + (["SSH_AGENT_PID", "SSH_AUTH_SOCK"] if not is_win else [])
        + ["SSL_CERT_FILE"]
        + (["SYSTEMDRIVE", "SYSTEMROOT", "TEMP"] if is_win else [])
        + (["TERM"] if stdout_is_atty else [])
        + (["TMP", "USERPROFILE"] if is_win else ["TMPDIR"])
        + ["VIRTUALENV_*"]
        + (["WINDIR"] if is_win else [])
        + ["http_proxy", "https_proxy", "no_proxy"]
    )
    assert pass_env == expected


def test_show_config_pkg_env_once(
    tox_project: ToxProjectCreator,
    patch_prev_py: Callable[[bool], tuple[str, str]],
) -> None:
    prev_ver, impl = patch_prev_py(True)
    ini = f"[tox]\nenv_list=py{prev_ver},py\n[testenv]\npackage=wheel"
    project = tox_project({"tox.ini": ini, "pyproject.toml": ""})
    result = project.run("c", "-e", "ALL", raise_on_config_fail=False)
    parser = ConfigParser(interpolation=None)
    parser.read_string(result.out)
    sections = set(parser.sections())
    assert sections == {"testenv:.pkg", f"testenv:.pkg-{impl}{prev_ver}", f"testenv:py{prev_ver}", "testenv:py", "tox"}


def test_show_config_pkg_env_skip(
    tox_project: ToxProjectCreator,
    patch_prev_py: Callable[[bool], tuple[str, str]],
) -> None:
    prev_ver, _impl = patch_prev_py(False)
    ini = f"[tox]\nenv_list=py{prev_ver},py\n[testenv]\npackage=wheel"
    project = tox_project({"tox.ini": ini, "pyproject.toml": ""})
    result = project.run("c", "-e", "ALL", raise_on_config_fail=False)
    parser = ConfigParser(interpolation=None)
    parser.read_string(result.out)
    sections = set(parser.sections())
    assert sections == {"testenv:.pkg", "tox", "testenv:py", f"testenv:py{prev_ver}"}


def test_show_config_select_only(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[tox]\nenv_list=\n a\n b", "pyproject.toml": ""})
    result = project.run("c", "-e", ".pkg,b,.pkg")
    result.assert_success()
    parser = ConfigParser(interpolation=None)
    parser.read_string(result.out)
    sections = list(parser.sections())
    assert sections == ["testenv:.pkg", "testenv:b"]


def test_show_config_alias(tox_project: ToxProjectCreator) -> None:
    outcome = tox_project({"tox.ini": ""}).run("c", "-e", "py", "-k", "setenv")
    outcome.assert_success()
    assert "set_env =" in outcome.out


def test_show_config_description_normalize(tox_project: ToxProjectCreator) -> None:
    tox_ini = "[testenv]\ndescription = A   magical\t pipe\n  of\tthis"
    outcome = tox_project({"tox.ini": tox_ini}).run("c", "-e", "py", "-k", "description")
    outcome.assert_success()
    assert outcome.out == "[testenv:py]\ndescription = A magical pipe of this\n"


def test_show_config_ini_comment_path(tox_project: ToxProjectCreator, tmp_path: Path) -> None:
    prj_path = tmp_path / "#magic"
    prj_path.mkdir()
    ini = """
    [testenv]
    package = skip
    set_env =
        A=1 # comment
        # more comment
    commands = {envpython} -c 'import os; print(os.linesep.join(f"{k}={v}" for k, v in os.environ.items()))'
    [testenv:py]
    set_env =
        {[testenv]set_env}
        B = {tox_root} # just some comment
    """
    project = tox_project({"tox.ini": dedent(ini)}, prj_path=prj_path)
    result = project.run("r", "-e", "py")
    result.assert_success()
    a_line = next(i for i in result.out.splitlines() if i.startswith("A="))  # pragma: no branch  # not found raises
    assert a_line == "A=1"
    b_line = next(i for i in result.out.splitlines() if i.startswith("B="))  # pragma: no branch  # not found raises
    assert b_line == f"B={prj_path}"


def test_show_config_cli_flag(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "", "pyproject.toml": ""})
    result = project.run("c", "-e", "py,.pkg", "-k", "package", "recreate", "--develop", "-r", "--no-recreate-pkg")
    expected = "[testenv:py]\npackage = editable\nrecreate = True\n\n[testenv:.pkg]\nrecreate = False\n"
    assert result.out == expected


def test_show_config_timeout_default(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv]\npakcage=skip"})
    result = project.run("c", "-e", "py", "-k", "suicide_timeout", "interrupt_timeout", "terminate_timeout")
    expected = "[testenv:py]\nsuicide_timeout = 0.0\ninterrupt_timeout = 0.3\nterminate_timeout = 0.2\n"
    assert result.out == expected


def test_show_config_timeout_custom(tox_project: ToxProjectCreator) -> None:
    ini = "[testenv]\npakcage=skip\nsuicide_timeout = 1\ninterrupt_timeout = 2.222\nterminate_timeout = 3.0\n"
    project = tox_project({"tox.ini": ini})
    result = project.run("c", "-e", "py", "-k", "suicide_timeout", "interrupt_timeout", "terminate_timeout")
    expected = "[testenv:py]\nsuicide_timeout = 1.0\ninterrupt_timeout = 2.222\nterminate_timeout = 3.0\n"
    assert result.out == expected


def test_show_config_help(tox_project: ToxProjectCreator) -> None:
    outcome = tox_project({"tox.ini": ""}).run("c", "-h")
    outcome.assert_success()


def test_show_config_core_host_python(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": ""})
    outcome = project.run("c", "--core", "-e", "py", "-k", "host_python")
    outcome.assert_success()
    assert f"host_python = {sys.executable}" in outcome.out


def test_show_config_matching_env_section(tox_project: ToxProjectCreator) -> None:
    ini = """
    [a]
    [testenv:a]
    deps = c>=1
    [testenv:b]
    deps = {[testenv:a]deps}"""
    project = tox_project({"tox.ini": ini})
    outcome = project.run("c", "-e", "a,b", "-k", "deps")
    outcome.assert_success()
    assert outcome.out.count("c>=1") == 2, outcome.out


def test_package_env_inherits_from_pkgenv(tox_project: ToxProjectCreator, demo_pkg_inline: Path) -> None:
    project = tox_project({"tox.ini": "[pkgenv]\npass_env = A, AA\ndeps=C\n D"})
    outcome = project.run(
        "c",
        "--root",
        str(demo_pkg_inline),
        "--workdir",
        str(project.path / ".tox"),
        "-k",
        "deps",
        "pass_env",
        "-e",
        "py,.pkg",
    )
    outcome.assert_success()
    exp = """
    [testenv:.pkg]
    deps =
      C
      D
    pass_env =
      A
      AA
    """
    exp = dedent(exp)
    assert exp in outcome.out


def test_cross_section_envpython_resolves_to_calling_env(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.ini": dedent("""\
            [testenv:a]
            package = skip
            commands = {envpython} -m black .

            [testenv:b]
            package = skip
            commands = {[testenv:a]commands}
        """),
    })
    outcome = project.run("c", "-e", "a,b", "-k", "commands")
    outcome.assert_success()
    lines = outcome.out.splitlines()
    a_cmd = next(line for line in lines if line.startswith("commands") and lines.index(line) < lines.index(""))
    b_cmd = [line for line in lines if line.startswith("commands")][-1]
    sep = os.sep
    assert f".tox{sep}a{sep}" in a_cmd
    assert f".tox{sep}b{sep}" in b_cmd
    assert f".tox{sep}a{sep}" not in b_cmd


def test_factor_conditional_falls_back_to_default(tox_project: ToxProjectCreator) -> None:
    py_ver = f"{sys.version_info[0]}.{sys.version_info[1]}"
    ini = f"""\
    [tox]
    env_list=py{py_ver}-a,py{py_ver}-b
    no_package=true
    [testenv]
    base_python =
        a: python{py_ver}
    """
    outcome = tox_project({"tox.ini": ini}).run("c", "-e", f"py{py_ver}-a,py{py_ver}-b", "-k", "base_python")
    outcome.assert_success()
    parser = ConfigParser(interpolation=None)
    parser.read_string(outcome.out)
    assert parser.get(f"testenv:py{py_ver}-a", "base_python") == f"python{py_ver}"
    assert parser.get(f"testenv:py{py_ver}-b", "base_python") == f"py{py_ver}"


def test_cross_section_factor_conditional_resolves_to_empty(tox_project: ToxProjectCreator) -> None:
    """Cross-section substitution resolves to empty when no factors match.

    Regression test for gh-3809: {[section]key} where key contains factor-conditional values should resolve to empty
    string when no factors match, not remain unresolved.

    """
    ini = """
    [tox]
    env_list=py39{,-keyfs_sqlite,-hash_hl}
    no_package=true
    [devpisettings]
    storagebackend=
        keyfs_sqlite: --backend-sqlite
    [testenv]
    commands=echo {[devpisettings]storagebackend}
    allowlist_externals = echo
    """
    outcome = tox_project({"tox.ini": ini}).run("c", "-e", "py39,py39-keyfs_sqlite", "-k", "commands")
    outcome.assert_success()
    parser = ConfigParser(interpolation=None)
    parser.read_string(outcome.out)
    assert parser.get("testenv:py39", "commands") == "echo"
    assert parser.get("testenv:py39-keyfs_sqlite", "commands") == "echo --backend-sqlite"


def test_core_on_platform(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true"})
    result = project.run("c", "-e", "py", "--core", "-k", "on_platform")
    result.assert_success()
    assert result.out == f"[testenv:py]\n\n[tox]\non_platform = {sys.platform}\n"



from __future__ import annotations

from argparse import Action, ArgumentParser, Namespace
from typing import TYPE_CHECKING
from unittest.mock import MagicMock

import pytest

from tox.report import HandledError
from tox.session.env_select import _env_completer, register_env_select_flags  # noqa: PLC2701

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture


@pytest.fixture
def completer_args() -> dict[str, Action | ArgumentParser | Namespace | str]:
    return {
        "prefix": "",
        "action": MagicMock(spec=Action),
        "parser": MagicMock(spec=ArgumentParser),
        "parsed_args": MagicMock(spec=Namespace),
    }


def test_env_completer_ini(tmp_path: Path, monkeypatch: pytest.MonkeyPatch, completer_args: dict[str, object]) -> None:
    config = tmp_path / "tox.ini"
    config.write_text("[tox]\nenv_list = py311,lint\n[testenv:docs]\n")
    monkeypatch.chdir(tmp_path)
    result = _env_completer(**completer_args)  # type: ignore[arg-type]
    assert "ALL" in result
    assert "py311" in result
    assert "lint" in result
    assert "docs" in result


def test_env_completer_toml(tmp_path: Path, monkeypatch: pytest.MonkeyPatch, completer_args: dict[str, object]) -> None:
    config = tmp_path / "tox.toml"
    config.write_text('[env.lint]\ncommands = [["ruff", "check"]]\n')
    monkeypatch.chdir(tmp_path)
    result = _env_completer(**completer_args)  # type: ignore[arg-type]
    assert "ALL" in result
    assert "lint" in result


def test_env_completer_no_config(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, completer_args: dict[str, object]
) -> None:
    monkeypatch.chdir(tmp_path)
    result = _env_completer(**completer_args)  # type: ignore[arg-type]
    assert result == ["ALL"]


def test_env_completer_handled_error(mocker: MockerFixture, completer_args: dict[str, object]) -> None:
    mocker.patch("tox.session.env_select.discover_source", side_effect=HandledError("bad config"))
    result = _env_completer(**completer_args)  # type: ignore[arg-type]
    assert result == []


def test_env_completer_includes_plugin_envs(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, mocker: MockerFixture, completer_args: dict[str, object]
) -> None:
    config = tmp_path / "tox.ini"
    config.write_text("[tox]\nenv_list = py311\n")
    monkeypatch.chdir(tmp_path)
    mocker.patch("tox.plugin.manager.Plugin.tox_extend_envs", return_value=[["plugin-env"]])
    result = _env_completer(**completer_args)  # type: ignore[arg-type]
    assert "ALL" in result
    assert "py311" in result
    assert "plugin-env" in result


def test_register_env_select_attaches_completer() -> None:
    parser = ArgumentParser()
    register_env_select_flags(parser, default=None, multiple=False)
    action = next(a for a in parser._actions if a.dest == "env")  # noqa: SLF001
    assert hasattr(action, "completer")
    assert action.completer is _env_completer



from __future__ import annotations

import sys

from tox.config.types import MissingRequiredConfigKeyError

if sys.version_info >= (3, 11):  # pragma: no cover (py311+)
    import tomllib
else:  # pragma: no cover (py311+)
    import tomli as tomllib


from typing import TYPE_CHECKING

from .ini import IniSource

if TYPE_CHECKING:
    from pathlib import Path


class LegacyToml(IniSource):
    FILENAME = "pyproject.toml"

    def __init__(self, path: Path) -> None:
        if path.name != self.FILENAME or not path.exists():
            raise ValueError
        with path.open("rb") as file_handler:
            toml_content = tomllib.load(file_handler)
        try:
            content = toml_content["tool"]["tox"]["legacy_tox_ini"]
        except KeyError as exc:
            msg = f"`tool.tox.legacy_tox_ini` missing from {path}"
            raise MissingRequiredConfigKeyError(msg) from exc
        super().__init__(path, content=content)


__all__ = ("LegacyToml",)



"""Expand TOML product dicts into environment name lists."""

from __future__ import annotations

from itertools import product
from typing import Any

from tox.config.loader.ini.factor import LATEST_PYTHON_MINOR_MAX, LATEST_PYTHON_MINOR_MIN


def expand_product(value: dict[str, Any]) -> list[str]:
    """Expand a product dict into a flat list of environment names.

    :param value: dict with ``product`` (list of factor groups) and optional ``exclude`` (list of env names to skip)

    :returns: list of environment names from the cartesian product of all factor groups, joined with ``-``

    """
    raw_groups = value["product"]
    if not isinstance(raw_groups, list):
        msg = f"product value must be a list of factor groups, got {type(raw_groups).__name__}"
        raise TypeError(msg)
    if not raw_groups:
        return []
    expanded = [expand_factor_group(g) for g in raw_groups]
    exclude = set(value.get("exclude") or [])
    return [name for combo in product(*expanded) if (name := "-".join(combo)) not in exclude]


def expand_factor_group(group: Any) -> list[str]:
    if isinstance(group, list):
        return [str(item) for item in group]
    if isinstance(group, dict) and "prefix" in group:
        return _expand_range(group)
    msg = f"factor group must be a list of strings or a range dict with 'prefix', got {type(group).__name__}"
    raise TypeError(msg)


def _expand_range(range_dict: dict[str, Any]) -> list[str]:
    prefix: str = str(range_dict["prefix"])
    has_start = "start" in range_dict
    has_stop = "stop" in range_dict
    if not has_start and not has_stop:
        msg = "range must have at least 'start' or 'stop'"
        raise TypeError(msg)
    start = range_dict.get("start", LATEST_PYTHON_MINOR_MIN)
    stop = range_dict.get("stop", LATEST_PYTHON_MINOR_MAX)
    if not isinstance(start, int):
        msg = f"range 'start' must be an integer, got {type(start).__name__}"
        raise TypeError(msg)
    if not isinstance(stop, int):
        msg = f"range 'stop' must be an integer, got {type(stop).__name__}"
        raise TypeError(msg)
    return [f"{prefix}{i}" for i in range(start, stop + 1)]


__all__ = [
    "expand_factor_group",
    "expand_product",
]



from __future__ import annotations

import sys
from textwrap import dedent
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable

    from tox.pytest import ToxProjectCreator


def test_depends_wheel(tox_project: ToxProjectCreator, patch_prev_py: Callable[[bool], tuple[str, str]]) -> None:
    prev_ver, impl = patch_prev_py(True)  # has previous python
    ver = sys.version_info[0:2]
    py = f"py{''.join(str(i) for i in ver)}"
    prev_py = f"py{prev_ver}"
    ini = f"""
    [tox]
    env_list = py,{py},{prev_py},py31,cov2,cov
    [testenv]
    package = wheel
    [testenv:cov]
    depends = py,{py},{prev_py},py31
    skip_install = true
    [testenv:cov2]
    depends = cov
    skip_install = true
    """
    project = tox_project({"tox.ini": ini, "pyproject.toml": ""})
    outcome = project.run("de")
    outcome.assert_success()

    expected = f"""
    Execution order: py, {py}, {prev_py}, py31, cov, cov2
    ALL
       py ~ .pkg
       {py} ~ .pkg
       {prev_py} ~ .pkg | .pkg-{impl}{prev_ver}
       py31 ~ .pkg
       cov2
          cov
             py ~ .pkg
             {py} ~ .pkg
             {prev_py} ~ .pkg | .pkg-{impl}{prev_ver}
             py31 ~ .pkg
       cov
          py ~ .pkg
          {py} ~ .pkg
          {prev_py} ~ .pkg | .pkg-{impl}{prev_ver}
          py31 ~ .pkg
    """
    assert outcome.out == dedent(expected).lstrip()


def test_depends_sdist(tox_project: ToxProjectCreator, patch_prev_py: Callable[[bool], tuple[str, str]]) -> None:
    prev_ver, _impl = patch_prev_py(True)  # has previous python
    ver = sys.version_info[0:2]
    py = f"py{''.join(str(i) for i in ver)}"
    prev_py = f"py{prev_ver}"
    ini = f"""
    [tox]
    env_list = py,{py},{prev_py},py31,cov2,cov
    [testenv]
    package = sdist
    [testenv:cov]
    depends = py,{py},{prev_py},py31
    skip_install = true
    [testenv:cov2]
    depends = cov
    skip_install = true
    """
    project = tox_project({"tox.ini": ini, "pyproject.toml": ""})
    outcome = project.run("de")
    outcome.assert_success()

    expected = f"""
    Execution order: py, {py}, {prev_py}, py31, cov, cov2
    ALL
       py ~ .pkg
       {py} ~ .pkg
       {prev_py} ~ .pkg
       py31 ~ .pkg
       cov2
          cov
             py ~ .pkg
             {py} ~ .pkg
             {prev_py} ~ .pkg
             py31 ~ .pkg
       cov
          py ~ .pkg
          {py} ~ .pkg
          {prev_py} ~ .pkg
          py31 ~ .pkg
    """
    assert outcome.out == dedent(expected).lstrip()


def test_depends_glob_star(tox_project: ToxProjectCreator) -> None:
    toml = """
    env_list = ["3.12", "3.13", "3.14", "lint", "cov"]
    [env_run_base]
    package = "skip"
    [env.cov]
    depends = ["3.*"]
    """
    project = tox_project({"tox.toml": toml})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: 3.12, 3.13, 3.14, lint, cov
    ALL
       3.12
       3.13
       3.14
       lint
       cov
          3.12
          3.13
          3.14
    """)
    assert outcome.out == expected


def test_depends_glob_question_mark(tox_project: ToxProjectCreator) -> None:
    toml = """
    env_list = ["a1", "a2", "ab", "cov"]
    [env_run_base]
    package = "skip"
    [env.cov]
    depends = ["a?"]
    """
    project = tox_project({"tox.toml": toml})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: a1, a2, ab, cov
    ALL
       a1
       a2
       ab
       cov
          a1
          a2
          ab
    """)
    assert outcome.out == expected


def test_depends_glob_no_match(tox_project: ToxProjectCreator) -> None:
    toml = """
    env_list = ["lint", "cov"]
    [env_run_base]
    package = "skip"
    [env.cov]
    depends = ["py*"]
    """
    project = tox_project({"tox.toml": toml})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: lint, cov
    ALL
       lint
       cov
    """)
    assert outcome.out == expected


def test_depends_glob_mixed(tox_project: ToxProjectCreator) -> None:
    toml = """
    env_list = ["3.13", "3.14", "lint", "cov"]
    [env_run_base]
    package = "skip"
    [env.cov]
    depends = ["lint", "3.*"]
    """
    project = tox_project({"tox.toml": toml})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: 3.13, 3.14, lint, cov
    ALL
       3.13
       3.14
       lint
       cov
          3.13
          3.14
          lint
    """)
    assert outcome.out == expected


def test_depends_glob_excludes_self(tox_project: ToxProjectCreator) -> None:
    toml = """
    env_list = ["a", "b", "cov"]
    [env_run_base]
    package = "skip"
    [env.cov]
    depends = ["*"]
    """
    project = tox_project({"tox.toml": toml})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: a, b, cov
    ALL
       a
       b
       cov
          a
          b
    """)
    assert outcome.out == expected


def test_depends_ini_glob_star(tox_project: ToxProjectCreator) -> None:
    ini = """
    [tox]
    env_list = 3.12,3.13,3.14,lint,cov
    [testenv]
    skip_install = true
    [testenv:cov]
    depends = 3.*
    """
    project = tox_project({"tox.ini": ini})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: 3.12, 3.13, 3.14, lint, cov
    ALL
       3.12
       3.13
       3.14
       lint
       cov
          3.12
          3.13
          3.14
    """)
    assert outcome.out == expected


def test_depends_ini_glob_question_mark(tox_project: ToxProjectCreator) -> None:
    ini = """
    [tox]
    env_list = a1,a2,ab,cov
    [testenv]
    skip_install = true
    [testenv:cov]
    depends = a?
    """
    project = tox_project({"tox.ini": ini})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: a1, a2, ab, cov
    ALL
       a1
       a2
       ab
       cov
          a1
          a2
          ab
    """)
    assert outcome.out == expected


def test_depends_ini_glob_no_match(tox_project: ToxProjectCreator) -> None:
    ini = """
    [tox]
    env_list = lint,cov
    [testenv]
    skip_install = true
    [testenv:cov]
    depends = py*
    """
    project = tox_project({"tox.ini": ini})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: lint, cov
    ALL
       lint
       cov
    """)
    assert outcome.out == expected


def test_depends_ini_glob_mixed(tox_project: ToxProjectCreator) -> None:
    ini = """
    [tox]
    env_list = 3.13,3.14,lint,cov
    [testenv]
    skip_install = true
    [testenv:cov]
    depends = lint,3.*
    """
    project = tox_project({"tox.ini": ini})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: 3.13, 3.14, lint, cov
    ALL
       3.13
       3.14
       lint
       cov
          3.13
          3.14
          lint
    """)
    assert outcome.out == expected


def test_depends_ini_glob_excludes_self(tox_project: ToxProjectCreator) -> None:
    ini = """
    [tox]
    env_list = a,b,cov
    [testenv]
    skip_install = true
    [testenv:cov]
    depends = *
    """
    project = tox_project({"tox.ini": ini})
    outcome = project.run("de")
    outcome.assert_success()
    expected = dedent("""\
    Execution order: a, b, cov
    ALL
       a
       b
       cov
          a
          b
    """)
    assert outcome.out == expected


def test_depends_help(tox_project: ToxProjectCreator) -> None:
    outcome = tox_project({"tox.ini": ""}).run("de", "-h")
    outcome.assert_success()



"""A reader that drains a stream via its file descriptor, following CPython's subprocess approach."""

from __future__ import annotations

from abc import ABC, abstractmethod
from threading import Event, Thread
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import sys
    from collections.abc import Callable
    from types import TracebackType

    if sys.version_info >= (3, 11):  # pragma: no cover (py311+)
        from typing import Self
    else:  # pragma: no cover (<py311)
        from typing_extensions import Self


WAIT_GENERAL = 0.05


class ReadViaThread(ABC):
    def __init__(self, file_no: int, handler: Callable[[bytes], int], name: str, drain: bool) -> None:  # noqa: FBT001
        self.file_no = file_no
        self.stop = Event()
        self.thread = Thread(target=self._read_stream, name=f"tox-r-{name}-{file_no}")
        self.handler = handler
        self._on_exit_drain = drain

    def __enter__(self) -> Self:
        self.thread.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.stop.set()
        while self.thread.is_alive():
            self.thread.join(WAIT_GENERAL)
        if self._on_exit_drain:
            self._drain_stream()

    @abstractmethod
    def _read_stream(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def _drain_stream(self) -> None:
        raise NotImplementedError



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from packaging.requirements import Requirement

from .virtual_env.package.util import dependencies_with_extras_from_markers

if TYPE_CHECKING:
    from pathlib import Path

if sys.version_info >= (3, 11):  # pragma: no cover (py311+)
    import tomllib
else:  # pragma: no cover (py311+)
    import tomli as tomllib


def resolve_extras_static(root: Path, extras: set[str]) -> list[Requirement] | None:
    pyproject_file = root / "pyproject.toml"
    if not pyproject_file.exists():
        return None
    with pyproject_file.open("rb") as file_handler:
        pyproject = tomllib.load(file_handler)
    if "project" not in pyproject:
        return None
    project = pyproject["project"]
    for dynamic in project.get("dynamic", []):
        if dynamic == "dependencies" or (extras and dynamic == "optional-dependencies"):
            return None
    deps_with_markers: list[tuple[Requirement, set[str | None]]] = [
        (Requirement(i), {None}) for i in project.get("dependencies", [])
    ]
    optional_deps = project.get("optional-dependencies", {})
    for extra, reqs in optional_deps.items():
        deps_with_markers.extend((Requirement(req), {extra}) for req in (reqs or []))
    return dependencies_with_extras_from_markers(
        deps_with_markers=deps_with_markers,
        extras=extras,
        package_name=project.get("name", "."),
        available_extras=set(optional_deps.keys()),
    )


__all__ = [
    "resolve_extras_static",
]



"""Contains the plugin manager object."""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

import pluggy

from tox import provision
from tox.config.loader import api as loader_api
from tox.session.cmd.run import parallel, sequential
from tox.tox_env import package as package_api
from tox.tox_env.python.virtual_env import runner
from tox.tox_env.python.virtual_env.package import cmd_builder, pyproject
from tox.tox_env.register import REGISTER, ToxEnvRegister

from . import NAME, spec
from .inline import load_inline

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path
    from types import ModuleType

    from tox.config.cli.parser import ToxParser
    from tox.config.sets import ConfigSet, EnvConfigSet
    from tox.execute import Outcome
    from tox.session.state import State
    from tox.tox_env.api import ToxEnv


class Plugin:
    def __init__(self) -> None:
        self.manager: pluggy.PluginManager = pluggy.PluginManager(NAME)
        self.manager.add_hookspecs(spec)
        self.inline_module: ModuleType | None = None

    def _register_plugins(self, inline: ModuleType | None) -> None:
        from tox.session import state  # noqa: PLC0415
        from tox.session.cmd import (  # noqa: PLC0415
            depends,
            devenv,
            exec_,
            legacy,
            list_env,
            man,
            quickstart,
            schema,
            show_config,
            version_flag,
        )

        self.inline_module = inline
        if inline is not None:
            self.manager.register(inline)
        self._load_external_plugins()
        internal_plugins = (
            loader_api,
            provision,
            runner,
            pyproject,
            cmd_builder,
            legacy,
            version_flag,
            exec_,
            quickstart,
            show_config,
            schema,
            devenv,
            list_env,
            man,
            depends,
            parallel,
            sequential,
            package_api,
        )
        for plugin in internal_plugins:
            self.manager.register(plugin)
        self.manager.register(state)
        try:
            self.manager.check_pending()
        except pluggy.PluginValidationError:
            if inline is None:
                raise
            logging.warning("toxfile.py uses hooks not available in this tox version, skipping inline plugin")
            self.manager.unregister(inline)
            self.inline_module = None
            self.manager.check_pending()

    def _load_external_plugins(self) -> None:
        for name in os.environ.get("TOX_DISABLED_EXTERNAL_PLUGINS", "").split(","):
            self.manager.set_blocked(name)
        self.manager.load_setuptools_entrypoints(NAME)

    def tox_extend_envs(self) -> list[Iterable[str]]:
        additional_env_names_hook_value = self.manager.hook.tox_extend_envs()
        # NOTE: S101 is suppressed below to allow for type narrowing in MyPy
        assert isinstance(additional_env_names_hook_value, list)  # noqa: S101
        return additional_env_names_hook_value

    def tox_add_option(self, parser: ToxParser) -> None:
        self.manager.hook.tox_add_option(parser=parser)

    def tox_add_core_config(self, core_conf: ConfigSet, state: State) -> None:
        self.manager.hook.tox_add_core_config(core_conf=core_conf, state=state)

    def tox_add_env_config(self, env_conf: EnvConfigSet, state: State) -> None:
        self.manager.hook.tox_add_env_config(env_conf=env_conf, state=state)

    def tox_register_tox_env(self, register: ToxEnvRegister) -> None:
        self.manager.hook.tox_register_tox_env(register=register)

    def tox_before_run_commands(self, tox_env: ToxEnv) -> None:
        self.manager.hook.tox_before_run_commands(tox_env=tox_env)

    def tox_after_run_commands(self, tox_env: ToxEnv, exit_code: int, outcomes: list[Outcome]) -> None:
        self.manager.hook.tox_after_run_commands(tox_env=tox_env, exit_code=exit_code, outcomes=outcomes)

    def tox_on_install(self, tox_env: ToxEnv, arguments: Any, section: str, of_type: str) -> None:
        self.manager.hook.tox_on_install(tox_env=tox_env, arguments=arguments, section=section, of_type=of_type)

    def tox_env_teardown(self, tox_env: ToxEnv) -> None:
        self.manager.hook.tox_env_teardown(tox_env=tox_env)

    def load_plugins(self, path: Path) -> None:
        for plugin in self.manager.get_plugins():  # make sure we start with a clean state, repeated in memory run
            self.manager.unregister(plugin)
        inline = _load_inline(path)
        self._register_plugins(inline)
        REGISTER._register_tox_env_types(self)  # noqa: SLF001


def _load_inline(path: Path) -> ModuleType | None:  # used to be able to unregister plugin tests
    return load_inline(path)


MANAGER = Plugin()

__all__ = (
    "MANAGER",
    "Plugin",
)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 11):  # pragma: no cover (py311+)
        from typing import Self
    else:  # pragma: no cover (<py311)
        from typing_extensions import Self


class Section:  # noqa: PLW1641
    """tox configuration section."""

    SEP = ":"  #: string used to separate the prefix and the section in the key

    def __init__(self, prefix: str | None, name: str) -> None:
        self._prefix = prefix
        self._name = name

    @classmethod
    def from_key(cls: type[Self], key: str) -> Self:
        """Create a section from a section key.

        :param key: the section key

        :returns: the constructed section

        """
        sep_at = key.find(cls.SEP)
        if sep_at == -1:
            prefix, name = None, key
        else:
            prefix, name = key[:sep_at], key[sep_at + 1 :]
        return cls(prefix, name)

    @property
    def prefix(self) -> str | None:
        """:returns: the prefix of the section"""
        return self._prefix

    @property
    def name(self) -> str:
        """:returns: the name of the section"""
        return self._name

    @property
    def key(self) -> str:
        """:returns: the section key"""
        return self.SEP.join(i for i in (self._prefix, self._name) if i is not None)

    def __str__(self) -> str:
        return self.key

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(prefix={self._prefix!r}, name={self._name!r})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and (self._prefix, self._name) == (
            other._prefix,
            other.name,
        )


__all__ = [
    "Section",
]



"""Expand tox factor expressions to tox environment list."""

from __future__ import annotations

import re
import sys
import sysconfig
from itertools import chain, groupby, product
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator

LATEST_PYTHON_MINOR_MIN: int = 10
LATEST_PYTHON_MINOR_MAX: int = 14


def filter_for_env(value: str, name: str | None) -> str:
    current = (
        set(chain.from_iterable([(i for i, _ in a) for a in find_factor_groups(name)])) if name is not None else set()
    )
    current.add(sys.platform)
    parts = sysconfig.get_platform().rsplit("-", 1)
    if len(parts) > 1:
        current.add(parts[-1])
    overall: list[str] = []
    active_continuation = False
    pending_skip = False
    for factors, content in expand_factors(value):
        if factors is None:
            if pending_skip and not active_continuation and not content.endswith("\\"):
                pending_skip = False
                continue
            if content:
                overall.append(content)
            active_continuation = content.endswith("\\") if content else active_continuation
            pending_skip = False
        else:
            matched = any(all((a_name in current) ^ negate for a_name, negate in group) for group in factors)
            if matched:
                overall.append(content)
                active_continuation = content.endswith("\\")
                pending_skip = False
            else:
                pending_skip = content.endswith("\\")
    return "\n".join(overall)


def find_envs(value: str) -> Iterator[str]:
    seen = set()
    for factors, _ in expand_factors(value):
        if factors is not None:
            for group in factors:
                env = explode_factor(group)
                if env not in seen:
                    yield env
                    seen.add(env)


def extend_factors(value: str) -> Iterator[str]:
    for group in find_factor_groups(value):
        yield explode_factor(group)


def explode_factor(group: list[tuple[str, bool]]) -> str:
    return "-".join([name for name, _ in group])


def expand_factors(value: str) -> Iterator[tuple[list[list[tuple[str, bool]]] | None, str]]:
    for line in value.split("\n"):
        factors: list[list[tuple[str, bool]]] | None = None
        marker_search = re.search(
            r"""
            :           # colon separator
            ( \s | $ )  # followed by whitespace or end of string
            """,
            line,
            re.VERBOSE,
        )
        marker_at, content = marker_search.start() if marker_search else -1, line
        if marker_at != -1:
            try:
                factors = list(find_factor_groups(line[:marker_at].strip()))
            except ValueError:
                pass  # when cannot extract factors keep the entire line
            else:
                content = line[marker_at + 1 :].strip()
        yield factors, content


def find_factor_groups(value: str) -> Iterator[list[tuple[str, bool]]]:
    """Transform '{py,!pi}-{a,b},c' to [{'py', 'a'}, {'py', 'b'}, {'pi', 'a'}, {'pi', 'b'}, {'c'}]."""
    value = expand_ranges(value)
    for env in expand_env_with_negation(value):
        yield [name_with_negate(f) for f in env.split("-")]


_FACTOR_RE = re.compile(
    r"""
    (?:
        !?              # optional negation prefix
        [\w.*?]         # first char: word char, dot, or glob wildcards
        [\w.*?-]*       # remaining chars: word chars, dots, glob wildcards, or hyphens
    |
        ^$              # or an empty string
    )
    """,
    re.VERBOSE,
)


def expand_env_with_negation(value: str) -> Iterator[str]:
    """Transform '{py,!pi}-{a,b},c' to ['py-a', 'py-b', '!pi-a', '!pi-b', 'c']."""
    for key, group in groupby(
        re.split(
            r"""
            ( (?: \{ [^}]+ \} )+ )  # one or more brace groups
            |                        # or
            ,                        # comma separator
            """,
            value,
            flags=re.VERBOSE,
        ),
        key=bool,
    ):
        if key:
            group_str = "".join(group).strip()
            elements = re.split(
                r"""
                \{          # opening brace
                ( [^}]+ )   # capture contents
                \}          # closing brace
                """,
                group_str,
                flags=re.VERBOSE,
            )
            parts = [[i.strip() for i in elem.split(",")] for elem in elements]
            for variant in product(*parts):
                variant_str = "".join(variant)
                if not all(_FACTOR_RE.fullmatch(i) for i in variant_str.split("-")):
                    raise ValueError(variant_str)
                yield variant_str


def name_with_negate(factor: str) -> tuple[str, bool]:
    negated = is_negated(factor)
    result = factor[1:] if negated else factor
    return result, negated


def is_negated(factor: str) -> bool:
    return factor.startswith("!")


def expand_ranges(value: str) -> str:
    """Expand ranges in env expressions.

    Supports closed ranges ``{10-13}``, right-open ``{10-}`` (upper bound = :data:`LATEST_PYTHON_MINOR_MAX`), and
    left-open ``{-13}`` (lower bound = :data:`LATEST_PYTHON_MINOR_MIN`).

    """
    matches = re.findall(
        r"""
        (                       # outer capture group
            ( \d+ ) - ( \d+ )   # closed range: start-end
            |
            ( \d+ ) -           # right-open range: start-
            |
            (?<= [{,] ) - ( \d+ )  # left-open range: -end (preceded by { or ,)
            |
            \d+                 # single number
        )
        (?: , | \} )            # followed by comma or closing brace
        """,
        value,
        re.VERBOSE,
    )
    for src, start_, end_, open_start, open_end in matches:
        if src and start_ and end_:
            start, end = int(start_), int(end_)
            direction = 1 if start < end else -1
            expansion = ",".join(str(x) for x in range(start, end + direction, direction))
            value = value.replace(src, expansion, 1)
        elif open_start:
            expansion = ",".join(str(x) for x in range(int(open_start), LATEST_PYTHON_MINOR_MAX + 1))
            value = value.replace(f"{open_start}-", expansion, 1)
        elif open_end:
            expansion = ",".join(str(x) for x in range(LATEST_PYTHON_MINOR_MIN, int(open_end) + 1))
            value = value.replace(f"-{open_end}", expansion, 1)
    return value


__all__ = (
    "LATEST_PYTHON_MINOR_MAX",
    "LATEST_PYTHON_MINOR_MIN",
    "expand_factors",
    "expand_ranges",
    "extend_factors",
    "filter_for_env",
    "find_envs",
)



"""Please keep this file Python 2.7 compatible. See https://tox.readthedocs.io/en/rewrite/development.html#code-style-guide"""

from __future__ import annotations

import os
import sys
import tarfile
from pathlib import Path
from textwrap import dedent
from zipfile import ZipFile

name = "demo_pkg_inline"
pkg_name = name.replace("_", "-")

version = "1.0.0"
dist_info = f"{name}-{version}.dist-info"
logic = f"{name}/__init__.py"
plugin = f"{name}/example_plugin.py"
entry_points = f"{dist_info}/entry_points.txt"
metadata = f"{dist_info}/METADATA"
wheel = f"{dist_info}/WHEEL"
record = f"{dist_info}/RECORD"
content = {
    logic: f"def do():\n    print('greetings from {name}')",
    plugin: """
        try:
            from tox.config.cli.parser import ToxParser
            from tox.plugin import impl
            from tox.tox_env.python.virtual_env.runner import VirtualEnvRunner
            from tox.tox_env.register import ToxEnvRegister
        except ImportError:
            pass
        else:
            class ExampleVirtualEnvRunner(VirtualEnvRunner):
                @staticmethod
                def id() -> str:
                    return "example"
            @impl
            def tox_register_tox_env(register: ToxEnvRegister) -> None:
                register.add_run_env(ExampleVirtualEnvRunner)
            @impl
            def tox_add_option(parser: ToxParser) -> None:
                parser.add_argument("--demo-plugin", action="store_true", help="demo plugin flag")
        """,
}
metadata_files = {
    entry_points: f"""
        [tox]
        example = {name}.example_plugin""",
    metadata: """
        Metadata-Version: 2.1
        Name: {}
        Version: {}
        Summary: UNKNOWN
        Home-page: UNKNOWN
        Author: UNKNOWN
        Author-email: UNKNOWN
        License: UNKNOWN
        {}
        Platform: UNKNOWN

        UNKNOWN
       """.format(
        pkg_name,
        version,
        "\n        ".join(os.environ.get("METADATA_EXTRA", "").split("\n")),
    ),
    wheel: f"""
        Wheel-Version: 1.0
        Generator: {name}-{version}
        Root-Is-Purelib: true
        Tag: py{sys.version_info[0]}-none-any
       """,
    f"{dist_info}/top_level.txt": name,
    record: f"""
        {name}/__init__.py,,
        {dist_info}/METADATA,,
        {dist_info}/WHEEL,,
        {dist_info}/top_level.txt,,
        {dist_info}/RECORD,,
       """,
}


def build_wheel(
    wheel_directory: str,
    config_settings: dict[str, str] | None = None,  # noqa: ARG001
    metadata_directory: str | None = None,
) -> str:
    base_name = f"{name}-{version}-py{sys.version_info[0]}-none-any.whl"
    path = Path(wheel_directory) / base_name
    with ZipFile(str(path), "w") as zip_file_handler:
        for arc_name, data in content.items():  # pragma: no branch
            zip_file_handler.writestr(arc_name, dedent(data).strip())
        if metadata_directory is not None:
            for sub_directory, _, filenames in os.walk(metadata_directory):
                for filename in filenames:
                    zip_file_handler.write(
                        str(Path(metadata_directory) / sub_directory / filename),
                        str(Path(sub_directory) / filename),
                    )
        else:
            for arc_name, data in metadata_files.items():
                zip_file_handler.writestr(arc_name, dedent(data).strip())
    print(f"created wheel {path}")  # noqa: T201
    return base_name


def get_requires_for_build_wheel(config_settings: dict[str, str] | None = None) -> list[str]:  # noqa: ARG001
    return []  # pragma: no cover # only executed in non-host pythons


if os.environ.get("BACKEND_HAS_EDITABLE"):

    def build_editable(
        wheel_directory: str,
        config_settings: dict[str, str] | None = None,
        metadata_directory: str | None = None,
    ) -> str:
        return build_wheel(wheel_directory, config_settings, metadata_directory)


def build_sdist(sdist_directory: str, config_settings: dict[str, str] | None = None) -> str:  # noqa: ARG001
    result = f"{name}-{version}.tar.gz"  # pragma: win32 cover
    with tarfile.open(str(Path(sdist_directory) / result), "w:gz") as tar:  # pragma: win32 cover
        root = Path(__file__).parent  # pragma: win32 cover
        tar.add(str(root / "build.py"), "build.py")  # pragma: win32 cover
        tar.add(str(root / "pyproject.toml"), "pyproject.toml")  # pragma: win32 cover
    return result  # pragma: win32 cover


def get_requires_for_build_sdist(config_settings: dict[str, str] | None = None) -> list[str]:  # noqa: ARG001
    return []  # pragma: no cover # only executed in non-host pythons



"""Defines how to run a single tox environment."""

from __future__ import annotations

import logging
import time
from typing import TYPE_CHECKING, NamedTuple, cast

from tox.execute.api import Outcome, StdinSource
from tox.report import HandledError
from tox.tox_env.errors import Fail, Skip
from tox.tox_env.python.virtual_env.package.pyproject import ToxBackendFailed

if TYPE_CHECKING:
    from pathlib import Path

    from tox.config.types import Command
    from tox.tox_env.api import ToxEnv
    from tox.tox_env.runner import RunToxEnv

LOGGER = logging.getLogger(__name__)


class ToxEnvRunResult(NamedTuple):
    name: str
    skipped: bool
    code: int
    outcomes: list[Outcome]
    duration: float
    ignore_outcome: bool = False
    fail_fast: bool = False
    unavailable: bool = False


def run_one(tox_env: RunToxEnv, no_test: bool, suspend_display: bool) -> ToxEnvRunResult:  # noqa: FBT001
    start_one = time.monotonic()
    name = tox_env.conf.name
    with tox_env.display_context(suspend_display):
        skipped, code, outcomes = _evaluate(tox_env, no_test)
    duration = time.monotonic() - start_one
    return ToxEnvRunResult(
        name, skipped, code, outcomes, duration, tox_env.conf["ignore_outcome"], tox_env.conf["fail_fast"]
    )


def _evaluate(tox_env: RunToxEnv, no_test: bool) -> tuple[bool, int, list[Outcome]]:  # noqa: FBT001
    skipped = False
    code: int = 0
    outcomes: list[Outcome] = []
    try:
        try:
            tox_env.setup()
            code, outcomes = run_commands(tox_env, no_test)
        except Skip as exception:
            LOGGER.warning("skipped because %s", exception)
            code = 0
            skipped = True
        except ToxBackendFailed as exception:
            LOGGER.error("%s", exception)  # noqa: TRY400
            raise SystemExit(exception.code)  # noqa: B904
        except Fail as exception:
            LOGGER.error("failed with %s", exception)  # noqa: TRY400
            code = 1
        except HandledError as exception:
            LOGGER.error("%s", exception)  # noqa: TRY400
            code = 1
        except Exception:  # pragma: no cover
            LOGGER.exception("internal error")  # pragma: no cover
            code = 2  # pragma: no cover
        finally:
            tox_env.teardown()
    except SystemExit as exception:  # setup command fails (interrupted or via invocation)
        code = cast("int", exception.code)
    return skipped, code, outcomes


def run_commands(tox_env: RunToxEnv, no_test: bool) -> tuple[int, list[Outcome]]:  # noqa: FBT001
    outcomes: list[Outcome] = []
    if no_test:
        exit_code = Outcome.OK
    else:
        from tox.plugin.manager import MANAGER  # importing this here to avoid circular import  # noqa: PLC0415

        chdir: Path = tox_env.conf["change_dir"]
        chdir.mkdir(exist_ok=True, parents=True)
        ignore_errors: bool = tox_env.conf["ignore_errors"]
        retry_count: int = tox_env.conf["commands_retry"]
        MANAGER.tox_before_run_commands(tox_env)
        status_pre, status_main, status_post = -1, -1, -1
        try:
            try:
                status_pre = run_command_set(tox_env, "commands_pre", chdir, ignore_errors, outcomes, retry_count)
                if status_pre == Outcome.OK or ignore_errors:
                    status_main = run_command_set(tox_env, "commands", chdir, ignore_errors, outcomes, retry_count)
                else:
                    status_main = Outcome.OK
            finally:
                status_post = run_command_set(tox_env, "commands_post", chdir, ignore_errors, outcomes, retry_count)
        finally:
            exit_code = status_pre or status_main or status_post  # first non-success
            MANAGER.tox_after_run_commands(tox_env, exit_code, outcomes)
    return exit_code, outcomes


def run_command_set(  # noqa: PLR0913
    tox_env: ToxEnv,
    key: str,
    cwd: Path,
    ignore_errors: bool,  # noqa: FBT001
    outcomes: list[Outcome],
    retry_count: int = 0,
) -> int:
    exit_code = Outcome.OK
    command_set: list[Command] = tox_env.conf[key]
    for at, cmd in enumerate(command_set):
        max_attempts = 1 if cmd.ignore_exit_code else retry_count + 1
        for attempt in range(1, max_attempts + 1):
            current_outcome = tox_env.execute(
                cmd.args,
                cwd=cwd,
                stdin=StdinSource.USER if getattr(tox_env.options, "no_capture", False) else StdinSource.user_only(),
                show=True,
                run_id=f"{key}[{at}]",
            )
            outcomes.append(current_outcome)
            try:
                if cmd.invert_exit_code:
                    current_outcome.assert_failure()
                else:
                    current_outcome.assert_success()
            except SystemExit as exception:
                if cmd.ignore_exit_code:
                    logging.warning("command failed but is marked ignore outcome so handling it as success")
                    break
                if attempt < max_attempts:
                    logging.warning(
                        "command failed (attempt %d of %d), retrying ...: %s",
                        attempt,
                        max_attempts,
                        cmd.shell,
                    )
                    continue
                if ignore_errors:
                    if exit_code == Outcome.OK:
                        exit_code = cast("int", exception.code)
                    break
                return cast("int", exception.code)
            else:
                break
    return exit_code


__all__ = (
    "ToxEnvRunResult",
    "run_command_set",
    "run_one",
)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from tox.config.loader.api import ConfigLoadArgs, Override
from tox.config.loader.ini import IniLoader
from tox.config.source.ini_section import IniSection

if TYPE_CHECKING:
    from collections.abc import Callable
    from configparser import ConfigParser


def test_ini_loader_keys(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    core = IniSection(None, "tox")
    loader = IniLoader(core, mk_ini_conf("\n[tox]\n\na=b\nc=d\n\n"), [], core_section=core)
    assert loader.found_keys() == {"a", "c"}


def test_ini_loader_repr(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    core = IniSection(None, "tox")
    loader = IniLoader(core, mk_ini_conf("\n[tox]\n\na=b\nc=d\n\n"), [Override("tox.a=1")], core_section=core)
    assert repr(loader) == "IniLoader(section=tox, overrides={'a': [Override('tox.a=1')]})"


def test_ini_loader_has_section(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    core = IniSection(None, "tox")
    loader = IniLoader(core, mk_ini_conf("[magic]\n[tox]\n\na=b\nc=d\n\n"), [], core_section=core)
    assert loader.get_section("magic") is not None


def test_ini_loader_has_no_section(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    core = IniSection(None, "tox")
    loader = IniLoader(core, mk_ini_conf("[tox]\n\na=b\nc=d\n\n"), [], core_section=core)
    assert loader.get_section("magic") is None


def test_ini_loader_raw(mk_ini_conf: Callable[[str], ConfigParser]) -> None:
    core = IniSection(None, "tox")
    args = ConfigLoadArgs([], "name", None)
    loader = IniLoader(core, mk_ini_conf("[tox]\na=b"), [], core_section=core)
    result = loader.load(key="a", of_type=str, conf=None, factory=None, args=args)
    assert result == "b"


@pytest.mark.parametrize("sep", ["\n", "\r\n"])
def test_ini_loader_raw_strip_escaped_newline(mk_ini_conf: Callable[[str], ConfigParser], sep: str) -> None:
    core = IniSection(None, "tox")
    args = ConfigLoadArgs([], "name", None)
    loader = IniLoader(core, mk_ini_conf(f"[tox]{sep}a=b\\{sep} c"), [], core_section=core)
    result = loader.load(key="a", of_type=str, conf=None, factory=None, args=args)
    assert result == "bc"


@pytest.mark.parametrize(
    ("case", "result"),
    [
        ("# a", ""),
        ("#", ""),
        ("a # w", "a"),
        ("a\t# w", "a"),
        ("a# w", "a"),
        ("a\\# w", "a# w"),
        ("#a\n b # w\n w", "b\nw"),
    ],
)
def test_ini_loader_strip_comments(mk_ini_conf: Callable[[str], ConfigParser], case: str, result: str) -> None:
    core = IniSection(None, "tox")
    args = ConfigLoadArgs([], "name", None)
    loader = IniLoader(core, mk_ini_conf(f"[tox]\na={case}"), [], core_section=core)
    outcome = loader.load(key="a", of_type=str, conf=None, factory=None, args=args)
    assert outcome == result



"""Execute that runs on local file system via subprocess-es."""

from __future__ import annotations

import fnmatch
import logging
import os
import shutil
import sys
from subprocess import DEVNULL, PIPE, TimeoutExpired
from typing import TYPE_CHECKING, Any

from tox.execute.api import Execute, ExecuteInstance, ExecuteOptions, ExecuteStatus
from tox.execute.request import ExecuteRequest, StdinSource
from tox.execute.util import shebang
from tox.tox_env.errors import Fail

if TYPE_CHECKING:
    import io
    from collections.abc import Generator, Sequence
    from types import TracebackType

    from tox.execute.stream import SyncWrite

# mypy: warn-unused-ignores=false

if sys.platform == "win32":  # explicit check for mypy # pragma: win32 cover
    # needs stdin/stdout handlers backed by overlapped IO
    if TYPE_CHECKING:  # the typeshed libraries don't contain this, so replace it with normal one
        from subprocess import Popen
    else:
        from asyncio.windows_utils import Popen
    from signal import CTRL_C_EVENT as SIG_INTERRUPT
    from signal import SIGTERM

    from .read_via_thread_windows import ReadViaThreadWindows as ReadViaThread

else:  # pragma: win32 no cover
    from signal import SIGINT as SIG_INTERRUPT
    from signal import SIGKILL, SIGTERM
    from subprocess import Popen

    from .read_via_thread_unix import ReadViaThreadUnix as ReadViaThread


IS_WIN = sys.platform == "win32"


class LocalSubProcessExecutor(Execute):
    def build_instance(  # noqa: PLR6301
        self,
        request: ExecuteRequest,
        options: ExecuteOptions,
        out: SyncWrite,
        err: SyncWrite,
    ) -> ExecuteInstance:
        return LocalSubProcessExecuteInstance(request, options, out, err)


class LocalSubprocessExecuteStatus(ExecuteStatus):
    def __init__(self, options: ExecuteOptions, out: SyncWrite, err: SyncWrite, process: Popen[bytes]) -> None:
        self._process: Popen[bytes] = process
        super().__init__(options, out, err)
        self._interrupted = False

    @property
    def exit_code(self) -> int | None:
        # need to poll here, to make sure the returncode we get is current
        self._process.poll()
        return self._process.returncode

    def interrupt(self) -> None:
        self._interrupted = True
        if self._process is not None:  # pragma: no branch
            # A three level stop mechanism for children - INT -> TERM -> KILL
            # communicate will wait for the app to stop, and then drain the standard streams and close them
            to_pid, host_pid = self._process.pid, os.getpid()
            msg = "requested interrupt of %d from %d, activate in %.2f"
            logging.warning(msg, to_pid, host_pid, self.options.suicide_timeout)
            if self.wait(self.options.suicide_timeout) is None:  # still alive -> INT
                # on Windows everyone in the same process group, so they got the message
                if sys.platform != "win32":  # pragma: win32 cover
                    msg = "send signal %s to %d from %d with timeout %.2f"
                    logging.warning(msg, f"SIGINT({SIG_INTERRUPT})", to_pid, host_pid, self.options.interrupt_timeout)
                    self._process.send_signal(SIG_INTERRUPT)
                if self.wait(self.options.interrupt_timeout) is None:  # still alive -> TERM # pragma: no branch
                    terminate_output = self.options.terminate_timeout
                    msg = "send signal %s to %d from %d with timeout %.2f"
                    logging.warning(msg, f"SIGTERM({SIGTERM})", to_pid, host_pid, terminate_output)
                    self._process.terminate()
                    # Windows terminate is UNIX kill
                    if sys.platform != "win32" and self.wait(terminate_output) is None:  # pragma: no branch
                        logging.warning(msg[:-18], f"SIGKILL({SIGKILL})", to_pid, host_pid)
                        self._process.kill()  # still alive -> KILL
                    self.wait()  # unconditional wait as kill should soon bring down the process
                logging.warning("interrupt finished with success")
            else:  # pragma: no cover # difficult to test, process must die just as it's being interrupted
                logging.warning("process already dead with %s within %s", self._process.returncode, host_pid)

    def wait(self, timeout: float | None = None) -> int | None:
        try:  # note wait in general might deadlock if output large, but we drain in background threads so not an issue
            return self._process.wait(timeout=timeout)
        except TimeoutExpired:
            return None

    def write_stdin(self, content: str) -> None:
        stdin = self._process.stdin
        if stdin is None:  # pragma: no branch
            return  # pragma: no cover
        bytes_content = content.encode()
        try:
            if sys.platform == "win32":  # explicit check for mypy  # pragma: win32 cover
                # on Windows we have a PipeHandle object here rather than a file stream
                import _overlapped  # noqa: PLC0415,PLC2701  # ty: ignore[unresolved-import] # no typeshed stubs

                ov = _overlapped.Overlapped(0)
                ov.WriteFile(stdin.handle, bytes_content)  # ty: ignore[unresolved-attribute] # PipeHandle on Windows
                result = ov.getresult(10)  # wait up to 10ms to perform the operation
                if result != len(bytes_content):
                    msg = f"failed to write to {stdin!r}"
                    raise RuntimeError(msg)
            else:
                stdin.write(bytes_content)
                stdin.flush()
        except OSError:  # pragma: no cover
            if self._interrupted:  # pragma: no cover
                pass  # pragma: no cover  # if the process was asked to exit in the meantime ignore write errors
            raise  # pragma: no cover

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pid={self._process.pid}, returncode={self._process.returncode!r})"

    @property
    def metadata(self) -> dict[str, Any]:
        return {"pid": self._process.pid} if self._process.pid else {}


class LocalSubprocessExecuteFailedStatus(ExecuteStatus):
    def __init__(self, options: ExecuteOptions, out: SyncWrite, err: SyncWrite, exit_code: int | None) -> None:
        super().__init__(options, out, err)
        self._exit_code = exit_code

    @property
    def exit_code(self) -> int | None:
        return self._exit_code

    def wait(self, timeout: float | None = None) -> int | None:  # noqa: ARG002
        return self._exit_code  # pragma: no cover

    def write_stdin(self, content: str) -> None:
        """Cannot write."""

    def interrupt(self) -> None:  # noqa: PLR6301
        return None  # pragma: no cover # nothing running so nothing to interrupt


class LocalSubProcessExecuteInstance(ExecuteInstance):
    def __init__(
        self,
        request: ExecuteRequest,
        options: ExecuteOptions,
        out: SyncWrite,
        err: SyncWrite,
        on_exit_drain: bool = True,  # noqa: FBT001, FBT002
    ) -> None:
        super().__init__(request, options, out, err)
        self.process: Popen[bytes] | None = None
        self._cmd: list[str] | None = None
        self._read_stderr: ReadViaThread | None = None
        self._read_stdout: ReadViaThread | None = None
        self._on_exit_drain = on_exit_drain

    @property
    def cmd(self) -> Sequence[str]:
        if self._cmd is None:
            base = self.request.cmd[0]
            executable = shutil.which(base, path=self.request.env["PATH"])
            if executable is None:
                cmd = self.request.cmd  # if failed to find leave as it is
            else:
                if self.request.allow is not None:
                    for allow in self.request.allow:
                        # 1. allow matches just the original name of the executable
                        # 2. allow matches the entire resolved path
                        if fnmatch.fnmatch(self.request.cmd[0], allow) or fnmatch.fnmatch(executable, allow):
                            break
                    else:
                        msg = f"{base} (resolves to {executable})" if base == executable else base
                        msg = f"{msg} is not allowed, use allowlist_externals to allow it"
                        raise Fail(msg)
                cmd = [executable]
                if sys.platform != "win32" and self.request.env.get("TOX_LIMITED_SHEBANG", "").strip():
                    shebang_line = shebang(executable)
                    if shebang_line:
                        cmd = [*shebang_line, executable]
                cmd.extend(self.request.cmd[1:])
            self._cmd = cmd
        return self._cmd

    def __enter__(self) -> ExecuteStatus:
        # adjust sub-process terminal size
        columns, lines = shutil.get_terminal_size(fallback=(-1, -1))
        if columns != -1:  # pragma: no branch
            self.request.env.setdefault("COLUMNS", str(columns))
        if lines != -1:  # pragma: no branch
            self.request.env.setdefault("LINES", str(lines))

        # --no-capture inherits console handles for interactive programs (e.g., Python REPL).
        # Allows terminal APIs to query console dimensions and interact with the terminal directly.
        inherit_console = self.options.no_capture
        stdout, stderr = self.get_stream_file_no("stdout"), self.get_stream_file_no("stderr")
        try:
            self.process = process = Popen(
                self.cmd,
                stdout=None if inherit_console else next(stdout),
                stderr=None if inherit_console else next(stderr),
                stdin={StdinSource.USER: None, StdinSource.OFF: DEVNULL, StdinSource.API: PIPE}[self.request.stdin],
                cwd=str(self.request.cwd),
                env=self.request.env,
            )
        except OSError as exception:
            # We log a nice error message to avout returning opaque error codes,
            # like exit code 2 (filenotfound).
            logging.error("Exception running subprocess %s", exception)  # noqa: TRY400
            return LocalSubprocessExecuteFailedStatus(self.options, self._out, self._err, exception.errno)

        status = LocalSubprocessExecuteStatus(self.options, self._out, self._err, process)
        if not inherit_console:
            drain, pid = self._on_exit_drain, self.process.pid
            self._read_stderr = ReadViaThread(stderr.send(process), self.err_handler, name=f"err-{pid}", drain=drain)
            self._read_stderr.__enter__()
            self._read_stdout = ReadViaThread(stdout.send(process), self.out_handler, name=f"out-{pid}", drain=drain)
            self._read_stdout.__enter__()
        return status

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        if self._read_stderr is not None:
            self._read_stderr.__exit__(exc_type, exc_val, exc_tb)
        if self._read_stdout is not None:
            self._read_stdout.__exit__(exc_type, exc_val, exc_tb)
        if self.process is not None:  # cleanup the file handlers
            for stream in (self.process.stdout, self.process.stderr, self.process.stdin):
                if stream is not None and not getattr(stream, "closed", False):
                    try:
                        stream.close()
                    except OSError as exc:  # pragma: no cover
                        logging.warning("error while trying to close %r with %r", stream, exc)  # pragma: no cover

    @staticmethod
    def get_stream_file_no(key: str) -> Generator[int, Popen[bytes], None]:
        allocated_pty = _pty(key)
        if allocated_pty is not None:
            main_fd, child_fd = allocated_pty
            yield child_fd
            os.close(child_fd)  # close the child process pipe
            yield main_fd
        else:
            process = yield PIPE
            stream = getattr(process, key)
            if sys.platform == "win32":  # explicit check for mypy # pragma: win32 cover
                yield stream.handle
            else:
                yield stream.name

    def set_out_err(self, out: SyncWrite, err: SyncWrite) -> tuple[SyncWrite, SyncWrite]:
        prev = self._out, self._err
        if self._read_stdout is not None:  # pragma: no branch
            self._read_stdout.handler = out.handler
        if self._read_stderr is not None:  # pragma: no branch
            self._read_stderr.handler = err.handler
        return prev


def _pty(key: str) -> tuple[int, int] | None:
    """Allocate a virtual terminal (pty) for a subprocess.

    A virtual terminal allows a process to perform syscalls that fetch attributes related to the tty, for example to
    determine whether to use colored output or enter interactive mode.

    The termios attributes of the controlling terminal stream will be copied to the allocated pty.

    :param key: The stream to copy attributes from. Either "stdout" or "stderr".

    :returns: (main_fd, child_fd) of an allocated pty; or None on error or if unsupported (win32).

    """
    if sys.platform == "win32":  # explicit check for mypy # pragma: win32 cover
        return None

    stream: io.TextIOWrapper = getattr(sys, key)

    # when our current stream is a tty, emulate pty for the child
    #   to allow host streams traits to be inherited
    if not stream.isatty():
        return None

    try:
        import fcntl  # noqa: PLC0415
        import pty  # noqa: PLC0415
        import struct  # noqa: PLC0415
        import termios  # noqa: PLC0415
    except ImportError:  # pragma: no cover
        return None  # cannot proceed on platforms without pty support

    try:
        main, child = pty.openpty()  # ty: ignore[possibly-missing-attribute] # Unix-only
    except OSError:  # could not open a tty
        return None  # pragma: no cover

    try:
        mode = termios.tcgetattr(stream)  # ty: ignore[possibly-missing-attribute] # Unix-only
        termios.tcsetattr(child, termios.TCSANOW, mode)  # ty: ignore[possibly-missing-attribute] # Unix-only
    except (termios.error, OSError):  # could not inherit traits  # ty: ignore[possibly-missing-attribute]
        return None  # pragma: no cover

    # adjust sub-process terminal size
    columns, lines = shutil.get_terminal_size(fallback=(-1, -1))
    if columns != -1 and lines != -1:
        size = struct.pack("HHHH", lines, columns, 0, 0)
        fcntl.ioctl(child, termios.TIOCSWINSZ, size)  # ty: ignore[possibly-missing-attribute] # Unix-only

    return main, child


__all__ = (
    "SIG_INTERRUPT",
    "LocalSubProcessExecuteInstance",
    "LocalSubProcessExecutor",
    "LocalSubprocessExecuteFailedStatus",
    "LocalSubprocessExecuteStatus",
)



from __future__ import annotations

import sys
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from tox.tox_env.errors import Fail
from tox.tox_env.python.pylock import Pylock

if TYPE_CHECKING:
    from pathlib import Path

    from tox.pytest import ToxProjectCreator

PYLOCK_TOML = dedent("""\
    lock-version = "1.0"
    created-by = "test-tool"

    [[packages]]
    name = "alpha"
    version = "1.0.0"

    [[packages.wheels]]
    url = "https://files.example.com/alpha-1.0.0-py3-none-any.whl"

    [packages.wheels.hashes]
    sha256 = "abc123"

    [[packages]]
    name = "beta"
    version = "2.0.0"

    [packages.sdist]
    url = "https://files.example.com/beta-2.0.0.tar.gz"

    [packages.sdist.hashes]
    sha256 = "def456"
""")


def test_pylock_install(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "pylock.toml"
            """,
            "pylock.toml": PYLOCK_TOML,
        },
    )
    execute_calls = project.patch_execute()
    result = project.run("r", "-e", "py")

    result.assert_success()

    req_file = str(project.path / ".tox" / "py" / "pylock.txt")
    found_calls = [(i[0][0].conf.name, i[0][3].run_id, i[0][3].cmd) for i in execute_calls.call_args_list]
    assert found_calls == [
        ("py", "install_pylock", ["python", "-I", "-m", "pip", "install", "--no-deps", "-r", req_file]),
    ]
    assert (project.path / ".tox" / "py" / "pylock.txt").read_text() == "alpha==1.0.0\nbeta==2.0.0"


def test_pylock_empty(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            """,
        },
    )
    execute_calls = project.patch_execute()
    result = project.run("r", "-e", "py")

    result.assert_success()
    assert not execute_calls.call_args_list


def test_pylock_not_found(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "missing.toml"
            """,
        },
    )
    result = project.run("r", "-e", "py")

    result.assert_failed()
    assert "pylock file 'missing.toml' not found" in result.out


def test_pylock_invalid_toml(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "pylock.toml"
            """,
            "pylock.toml": dedent("""\
                lock-version = "1.0"
                created-by = "test-tool"
                packages = "not-a-list"
            """),
        },
    )
    result = project.run("r", "-e", "py")

    result.assert_failed()
    assert "invalid pylock file" in result.out


def test_pylock_mutually_exclusive_with_deps(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            deps = ["pytest"]
            pylock = "pylock.toml"
            """,
            "pylock.toml": PYLOCK_TOML,
        },
    )
    result = project.run("r", "-e", "py")

    result.assert_failed()
    assert "cannot use both 'deps' and 'pylock'" in result.out


def test_pylock_recreate_on_change(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "pylock.toml"
            """,
            "pylock.toml": PYLOCK_TOML,
        },
    )
    execute_calls = project.patch_execute()
    result = project.run("r", "-e", "py")
    result.assert_success()

    (project.path / "pylock.toml").write_text(
        dedent("""\
            lock-version = "1.0"
            created-by = "test-tool"

            [[packages]]
            name = "alpha"
            version = "1.0.0"

            [[packages.wheels]]
            url = "https://files.example.com/alpha-1.0.0-py3-none-any.whl"

            [packages.wheels.hashes]
            sha256 = "abc123"
        """),
    )
    execute_calls.reset_mock()
    result_second = project.run("r", "-e", "py")
    result_second.assert_success()
    assert "py: recreate env because" in result_second.out


def test_pylock_no_reinstall_on_rerun(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "pylock.toml"
            """,
            "pylock.toml": PYLOCK_TOML,
        },
    )
    execute_calls = project.patch_execute()
    result = project.run("r", "-e", "py")
    result.assert_success()
    assert len(execute_calls.call_args_list) == 1

    execute_calls.reset_mock()
    result_second = project.run("r", "-e", "py")
    result_second.assert_success()
    assert not execute_calls.call_args_list


def test_pylock_filters_by_extras(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "pylock.toml"
            extras = ["docs"]
            """,
            "pylock.toml": dedent("""\
                lock-version = "1.0"
                created-by = "test-tool"
                extras = ["docs", "test"]

                [[packages]]
                name = "alpha"
                version = "1.0.0"

                [[packages.wheels]]
                url = "https://files.example.com/alpha-1.0.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "abc"

                [[packages]]
                name = "sphinx"
                version = "7.0.0"
                marker = "'docs' in extras"

                [[packages.wheels]]
                url = "https://files.example.com/sphinx-7.0.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "def"

                [[packages]]
                name = "pytest"
                version = "8.0.0"
                marker = "'test' in extras"

                [[packages.wheels]]
                url = "https://files.example.com/pytest-8.0.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "ghi"
            """),
        },
    )
    project.patch_execute()
    result = project.run("r", "-e", "py")

    result.assert_success()
    assert (project.path / ".tox" / "py" / "pylock.txt").read_text() == "alpha==1.0.0\nsphinx==7.0.0"


def test_pylock_filters_by_dependency_groups(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "pylock.toml"
            dependency_groups = ["dev"]
            """,
            "pylock.toml": dedent("""\
                lock-version = "1.0"
                created-by = "test-tool"
                dependency-groups = ["dev", "ci"]

                [[packages]]
                name = "alpha"
                version = "1.0.0"

                [[packages.wheels]]
                url = "https://files.example.com/alpha-1.0.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "abc"

                [[packages]]
                name = "ruff"
                version = "0.5.0"
                marker = "'dev' in dependency_groups"

                [[packages.wheels]]
                url = "https://files.example.com/ruff-0.5.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "def"

                [[packages]]
                name = "coverage"
                version = "7.0.0"
                marker = "'ci' in dependency_groups"

                [[packages.wheels]]
                url = "https://files.example.com/coverage-7.0.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "ghi"
            """),
        },
    )
    project.patch_execute()
    result = project.run("r", "-e", "py")

    result.assert_success()
    assert (project.path / ".tox" / "py" / "pylock.txt").read_text() == "alpha==1.0.0\nruff==0.5.0"


def test_pylock_filters_by_platform_marker(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "pylock.toml"
            """,
            "pylock.toml": dedent("""\
                lock-version = "1.0"
                created-by = "test-tool"

                [[packages]]
                name = "alpha"
                version = "1.0.0"

                [[packages.wheels]]
                url = "https://files.example.com/alpha-1.0.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "abc"

                [[packages]]
                name = "linuxonly"
                version = "1.0.0"
                marker = "sys_platform == 'linux'"

                [[packages.wheels]]
                url = "https://files.example.com/linuxonly-1.0.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "def"

                [[packages]]
                name = "winonly"
                version = "1.0.0"
                marker = "sys_platform == 'win32'"

                [[packages.wheels]]
                url = "https://files.example.com/winonly-1.0.0-py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "ghi"
            """),
        },
    )
    project.patch_execute()
    result = project.run("r", "-e", "py")

    result.assert_success()
    content = (project.path / ".tox" / "py" / "pylock.txt").read_text()
    assert "alpha==1.0.0" in content
    if sys.platform == "win32":
        assert "winonly==1.0.0" in content
        assert "linuxonly" not in content
    elif sys.platform == "linux":
        assert "linuxonly==1.0.0" in content
        assert "winonly" not in content
    else:
        assert "linuxonly" not in content
        assert "winonly" not in content


def test_pylock_requirements(tmp_path: Path) -> None:
    lock_file = tmp_path / "pylock.toml"
    lock_file.write_text(PYLOCK_TOML)
    pylock = Pylock(path=lock_file)

    result = [str(r) for r in pylock.requirements()]

    assert result == ["alpha==1.0.0", "beta==2.0.0"]


def test_pylock_requirements_filters_extras(tmp_path: Path) -> None:
    lock_file = tmp_path / "pylock.toml"
    lock_file.write_text(
        dedent("""\
        lock-version = "1.0"
        created-by = "test-tool"
        extras = ["docs"]

        [[packages]]
        name = "alpha"
        version = "1.0.0"

        [[packages.wheels]]
        url = "https://files.example.com/alpha-1.0.0-py3-none-any.whl"

        [packages.wheels.hashes]
        sha256 = "abc"

        [[packages]]
        name = "sphinx"
        version = "7.0.0"
        marker = "'docs' in extras"

        [[packages.wheels]]
        url = "https://files.example.com/sphinx-7.0.0-py3-none-any.whl"

        [packages.wheels.hashes]
        sha256 = "def"
    """),
    )

    with_extras = [str(r) for r in Pylock(path=lock_file, extras=frozenset({"docs"})).requirements()]
    assert with_extras == ["alpha==1.0.0", "sphinx==7.0.0"]

    without_extras = [str(r) for r in Pylock(path=lock_file).requirements()]
    assert without_extras == ["alpha==1.0.0"]


def test_pylock_requirements_invalid(tmp_path: Path) -> None:
    lock_file = tmp_path / "pylock.toml"
    lock_file.write_text(
        dedent("""\
        lock-version = "1.0"
        created-by = "test-tool"
        packages = "not-a-list"
    """),
    )
    pylock = Pylock(path=lock_file)

    with pytest.raises(Fail, match="invalid pylock file"):
        pylock.requirements()


@pytest.mark.slow
@pytest.mark.integration
def test_pylock_install_integration(
    tox_project: ToxProjectCreator,
    enable_pip_pypi_access: str | None,  # noqa: ARG001
) -> None:
    project = tox_project(
        {
            "tox.toml": """
            [env_run_base]
            skip_install = true
            pylock = "pylock.toml"
            commands = [["python", "-c", "import six; print(six.__version__)"]]
            """,
            "pylock.toml": dedent("""\
                lock-version = "1.0"
                created-by = "test-tool"

                [[packages]]
                name = "six"
                version = "1.17.0"

                [[packages.wheels]]
                url = "https://files.pythonhosted.org/packages/b7/ce/149a00dd41f10bc29e5921b496af8b574d8413afcd5e30f4c7e48bb4cb87/six-1.17.0-py2.py3-none-any.whl"

                [packages.wheels.hashes]
                sha256 = "4721f391ed90541fddacab5acf947aa0d3dc7d27b2e1e8ez91c67d35ad282f3"
            """),
        },
    )
    result = project.run("r", "-e", "py")

    result.assert_success()
    assert "1.17.0" in result.out



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import TypeAlias

TomlTypes: TypeAlias = dict[str, "TomlTypes"] | list["TomlTypes"] | str | int | float | bool | None

__all__ = [
    "TomlTypes",
]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from tox.tox_env.api import redact_value

if TYPE_CHECKING:
    from pathlib import Path

    from tox.pytest import ToxProjectCreator


def test_ensure_temp_dir_exists(tox_project: ToxProjectCreator) -> None:
    ini = "[testenv]\ncommands=python -c 'import os; os.path.exists(r\"{temp_dir}\")'"
    project = tox_project({"tox.ini": ini})
    result = project.run()
    result.assert_success()


def test_dont_cleanup_temp_dir(tox_project: ToxProjectCreator, tmp_path: Path) -> None:
    (tmp_path / "foo" / "bar").mkdir(parents=True)
    project = tox_project({"tox.ini": "[tox]\ntemp_dir=foo"})
    result = project.run()
    result.assert_success()
    assert (tmp_path / "foo" / "bar").exists()


def test_setenv_section_substitution(tox_project: ToxProjectCreator) -> None:
    ini = """[variables]
    var = VAR = val
    [testenv]
    setenv = {[variables]var}
    commands = python -c 'import os; os.environ["VAR"]'"""
    project = tox_project({"tox.ini": ini})
    result = project.run()
    result.assert_success()


@pytest.mark.parametrize(
    ("key", "do_redact"),
    [
        pytest.param("SOME_KEY", True, id="key"),
        pytest.param("API_FOO", True, id="api"),
        pytest.param("AUTH", True, id="auth"),
        pytest.param("CLIENT", True, id="client"),
        pytest.param("DB_PASSWORD", True, id="password"),
        pytest.param("FOO", False, id="foo"),
        pytest.param("GITHUB_TOKEN", True, id="token"),
        pytest.param("NORMAL_VAR", False, id="other"),
        pytest.param("S_PASSWD", True, id="passwd"),
        pytest.param("SECRET", True, id="secret"),
        pytest.param("SOME_ACCESS", True, id="access"),
        pytest.param("MY_CRED", True, id="cred"),
        pytest.param("MY_PRIVATE", True, id="private"),
        pytest.param("MY_PWD", True, id="pwd"),
    ],
)
def test_redact(key: str, do_redact: bool) -> None:
    """Ensures that redact_value works as expected."""
    result = redact_value(key, "foo")
    if do_redact:
        assert result == "***"
    else:
        assert result == "foo"
